"""Various classes that perform a transformation on instance sets."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.references import EntityReference, MetricReference, SemanticModelReference
from dbt_semantic_interfaces.type_enums import MetricType
from dbt_semantic_interfaces.type_enums.aggregation_type import AggregationType
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from metricflow_semantics.aggregation_properties import AggregationState
from metricflow_semantics.assert_one_arg import assert_exactly_one_arg_set
from metricflow_semantics.instances import (
    DimensionInstance,
    EntityInstance,
    GroupByMetricInstance,
    InstanceSet,
    InstanceSetTransform,
    MdoInstance,
    MetadataInstance,
    MetricInstance,
    SimpleMetricInputInstance,
    TimeDimensionInstance,
)
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.metric_lookup import MetricLookup
from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.spec_set import InstanceSpecSet
from metricflow_semantics.sql.sql_exprs import (
    SqlAggregateFunctionExpression,
    SqlColumnReference,
    SqlColumnReferenceExpression,
    SqlExpressionNode,
    SqlFunction,
    SqlFunctionExpression,
    SqlStringExpression,
)
from more_itertools import bucket
from typing_extensions import TypeVar

from metricflow.dataflow.builder.aggregation_helper import InstanceAliasMapping, NullFillValueMapping
from metricflow.dataflow.nodes.join_to_base import ValidityWindowJoinDescription
from metricflow.plan_conversion.select_column_gen import SelectColumnSet
from metricflow.sql.sql_plan import (
    SqlSelectColumn,
)

logger = logging.getLogger(__name__)

InstanceT = TypeVar("InstanceT", bound=MdoInstance)


@dataclass(frozen=True)
class _DimensionValidityParams:
    """Helper dataclass for managing dimension validity properties."""

    dimension_name: str
    time_granularity: TimeGranularity
    date_part: Optional[DatePart] = None


class CreateValidityWindowJoinDescription(InstanceSetTransform[Optional[ValidityWindowJoinDescription]]):
    """Create and return a ValidityWindowJoinDescription based on the given InstanceSet.

    During join resolution we need to determine whether or not a given data set represents a
    Type II SCD dataset - i.e., one with a validity window defined on each row. This requires
    checking the set of dimension instances and determining whether or not those originate from
    an SCD source, and extracting validity window information accordingly.
    """

    def __init__(self, semantic_model_lookup: SemanticModelLookup) -> None:
        """Initializer. The SemanticModelLookup is needed for getting the original model definition."""
        self._semantic_model_lookup = semantic_model_lookup

    def _get_validity_window_dimensions_for_semantic_model(
        self, semantic_model_reference: SemanticModelReference
    ) -> Optional[Tuple[_DimensionValidityParams, _DimensionValidityParams]]:
        """Returns a 2-tuple (start, end) of validity window dimensions info, if any exist in the semantic model."""
        semantic_model = self._semantic_model_lookup.get_by_reference(semantic_model_reference)
        assert semantic_model, f"Could not find semantic model {semantic_model_reference} after data set conversion!"

        start_dim = semantic_model.validity_start_dimension
        end_dim = semantic_model.validity_end_dimension

        # We do this instead of relying on has_validity_dimensions because this also does type refinement
        if not start_dim or not end_dim:
            return None

        assert start_dim.type_params, "Typechecker hint - validity info cannot exist without type params"
        assert end_dim.type_params, "Typechecker hint - validity info cannot exist without type params"

        return (
            _DimensionValidityParams(
                dimension_name=start_dim.name, time_granularity=start_dim.type_params.time_granularity
            ),
            _DimensionValidityParams(
                dimension_name=end_dim.name, time_granularity=end_dim.type_params.time_granularity
            ),
        )

    def transform(self, instance_set: InstanceSet) -> Optional[ValidityWindowJoinDescription]:
        """Find the Time Dimension specs defining a validity window, if any, and return it.

        This currently throws an exception if more than one such window is found, and effectively prevents
        us from processing a dataset composed of a join between two SCD semantic models. This restriction is in
        place as a temporary simplification - if there is need for this feature we can enable it.
        """
        semantic_model_to_window: Dict[SemanticModelReference, ValidityWindowJoinDescription] = {}
        instances_by_semantic_model = bucket(
            instance_set.time_dimension_instances, lambda x: x.origin_semantic_model_reference.semantic_model_reference
        )
        for semantic_model_reference in instances_by_semantic_model:
            validity_dims = self._get_validity_window_dimensions_for_semantic_model(semantic_model_reference)
            if validity_dims is None:
                continue

            start_dim, end_dim = validity_dims
            specs = {instance.spec for instance in instances_by_semantic_model[semantic_model_reference]}
            start_specs = [
                spec
                for spec in specs
                if spec.element_name == start_dim.dimension_name
                # TODO: [custom_granularity] - support custom granularities for SCDs. Note this requires
                # addition of SCD support for window sub-selection, similar to what we do for cumulative metrics
                # when we group by a different time grain (e.g., select last_value from window, etc.)
                and not spec.has_custom_grain
                and spec.base_granularity == start_dim.time_granularity
                and spec.date_part == start_dim.date_part
            ]
            end_specs = [
                spec
                for spec in specs
                if spec.element_name == end_dim.dimension_name
                # TODO: [custom_granularity] - support custom granularities for SCDs. Note this requires
                # addition of SCD support for window sub-selection, similar to what we do for cumulative metrics
                # when we group by a different time grain (e.g., select last_value from window, etc.)
                and not spec.has_custom_grain
                and spec.base_granularity == end_dim.time_granularity
                and spec.date_part == end_dim.date_part
            ]
            linkless_start_specs = {spec.without_entity_links for spec in start_specs}
            linkless_end_specs = {spec.without_entity_links for spec in end_specs}
            assert len(linkless_start_specs) == 1 and len(linkless_end_specs) == 1, (
                f"Did not find exactly one pair of specs from semantic model `{semantic_model_reference}` matching the validity "
                f"window end points defined in the semantic model. This means we cannot process an SCD join, because we "
                f"require exactly one validity window to be specified for the query! The window in the semantic model "
                f"is defined by start dimension `{start_dim}` and end dimension `{end_dim}`. We found "
                f"{len(linkless_start_specs)} linkless specs for window start ({linkless_start_specs}) and "
                f"{len(linkless_end_specs)} linkless specs for window end ({linkless_end_specs})."
            )
            # SCD join targets are joined as dimension links in much the same was as partitions are joined. Therefore,
            # we treat this like a partition time column join and take the dimension spec with the shortest set of
            # entity links so that the subquery uses the correct reference in the ON statement
            start_specs = sorted(start_specs, key=lambda x: len(x.entity_links))
            end_specs = sorted(end_specs, key=lambda x: len(x.entity_links))
            semantic_model_to_window[semantic_model_reference] = ValidityWindowJoinDescription(
                window_start_dimension=start_specs[0], window_end_dimension=end_specs[0]
            )

        assert len(semantic_model_to_window) < 2, (
            f"Found more than 1 set of validity window specs in the input instance set. This is not currently "
            f"supported, as joins between SCD semantic models are not yet allowed! {semantic_model_to_window}"
        )

        if semantic_model_to_window:
            return list(semantic_model_to_window.values())[0]

        return None


class FilterLinkableInstancesWithLeadingLink(InstanceSetTransform[InstanceSet]):
    """Return an instance set with the elements that have a specified leading link removed.

    e.g. Remove "listing__country" if the specified link is "listing".
    """

    def __init__(self, entity_link: EntityReference) -> None:
        """Remove elements with this link as the first element in "entity_links"."""
        self._entity_link = entity_link

    def _should_pass(self, linkable_spec: LinkableInstanceSpec) -> bool:
        if len(linkable_spec.entity_links) == 0:
            return not linkable_spec.reference == self._entity_link
        return linkable_spec.entity_links[0] != self._entity_link

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        # Normal to not filter anything if the instance set has no instances with links.
        filtered_dimension_instances = tuple(x for x in instance_set.dimension_instances if self._should_pass(x.spec))
        filtered_time_dimension_instances = tuple(
            x for x in instance_set.time_dimension_instances if self._should_pass(x.spec)
        )
        filtered_entity_instances = tuple(x for x in instance_set.entity_instances if self._should_pass(x.spec))
        filtered_group_by_metric_instances = tuple(
            x for x in instance_set.group_by_metric_instances if self._should_pass(x.spec)
        )

        output = InstanceSet(
            simple_metric_input_instances=instance_set.simple_metric_input_instances,
            dimension_instances=filtered_dimension_instances,
            time_dimension_instances=filtered_time_dimension_instances,
            entity_instances=filtered_entity_instances,
            group_by_metric_instances=filtered_group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )
        return output


class FilterElements(InstanceSetTransform[InstanceSet]):
    """Return an instance set with the elements that don't match any of the pass specs removed."""

    def __init__(
        self,
        include_specs: Optional[InstanceSpecSet] = None,
        exclude_specs: Optional[InstanceSpecSet] = None,
    ) -> None:
        """Constructor.

        Args:
            include_specs: If specified, pass only instances matching these specs.
            exclude_specs: If specified, pass only instances not matching these specs.
        """
        assert_exactly_one_arg_set(include_specs=include_specs, exclude_specs=exclude_specs)
        self._include_specs = include_specs
        self._exclude_specs = exclude_specs

    def _should_pass(self, element_spec: InstanceSpec) -> bool:
        # TODO: Use better matching function
        if self._include_specs:
            return any(x == element_spec for x in self._include_specs.all_specs)
        elif self._exclude_specs:
            return not any(x == element_spec for x in self._exclude_specs.all_specs)
        assert False

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        # Sanity check to make sure the specs are in the instance set
        available_specs = instance_set.spec_set.all_specs
        if self._include_specs:
            include_specs_not_found = []
            for include_spec in self._include_specs.all_specs:
                if include_spec not in available_specs:
                    include_specs_not_found.append(include_spec)
            if include_specs_not_found:
                raise RuntimeError(
                    LazyFormat(
                        "Include specs are not in the spec set - check if this node was constructed correctly.",
                        include_specs_not_found=include_specs_not_found,
                        available_specs=available_specs,
                    )
                )
        elif self._exclude_specs:
            exclude_specs_not_found = []

            for exclude_spec in self._exclude_specs.all_specs:
                if exclude_spec not in available_specs:
                    exclude_specs_not_found.append(exclude_spec)
            if exclude_specs_not_found:
                raise RuntimeError(
                    LazyFormat(
                        "Exclude specs are not in the spec set - check if this node was constructed correctly.",
                        exclude_specs_not_found=exclude_specs_not_found,
                        available_specs=available_specs,
                    )
                )
        else:
            assert False, "Include specs or exclude specs should have been specified."

        output = InstanceSet(
            simple_metric_input_instances=tuple(
                x for x in instance_set.simple_metric_input_instances if self._should_pass(x.spec)
            ),
            dimension_instances=tuple(x for x in instance_set.dimension_instances if self._should_pass(x.spec)),
            time_dimension_instances=tuple(
                x for x in instance_set.time_dimension_instances if self._should_pass(x.spec)
            ),
            entity_instances=tuple(x for x in instance_set.entity_instances if self._should_pass(x.spec)),
            group_by_metric_instances=tuple(
                x for x in instance_set.group_by_metric_instances if self._should_pass(x.spec)
            ),
            metric_instances=tuple(x for x in instance_set.metric_instances if self._should_pass(x.spec)),
            metadata_instances=tuple(x for x in instance_set.metadata_instances if self._should_pass(x.spec)),
        )
        return output


class ChangeSimpleMetricInputAggregationState(InstanceSetTransform[InstanceSet]):
    """Returns a new instance set where all simple-metric inputs are set as a different aggregation state."""

    def __init__(self, aggregation_state_changes: Dict[AggregationState, AggregationState]) -> None:
        """Constructor.

        Args:
            aggregation_state_changes: key is old aggregation state, value is the new aggregation state.
        """
        self._aggregation_state_changes = aggregation_state_changes

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        for instance in instance_set.simple_metric_input_instances:
            assert instance.aggregation_state in self._aggregation_state_changes, (
                f"Aggregation state: {instance.aggregation_state} not handled in change dict: "
                f"{self._aggregation_state_changes}"
            )

        # Copy the simple-metric inputs, but just change the aggregation state to COMPLETE.
        instances = tuple(
            SimpleMetricInputInstance(
                associated_columns=x.associated_columns,
                defined_from=x.defined_from,
                aggregation_state=self._aggregation_state_changes[x.aggregation_state],
                spec=x.spec,
            )
            for x in instance_set.simple_metric_input_instances
        )
        return InstanceSet(
            simple_metric_input_instances=instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )


class UpdateSimpleMetricInputFillNullsWith(InstanceSetTransform[InstanceSet]):
    """Returns a new instance set where all simple-metric inputs have been assigned the fill nulls with property."""

    def __init__(self, null_fill_value_mapping: NullFillValueMapping):
        """Initializer stores the input specs, which contain the fill_nulls_with for each simple-metric input."""
        self._null_fill_value_mapping = null_fill_value_mapping

    def _update_fill_nulls_with(
        self, instances: Tuple[SimpleMetricInputInstance, ...]
    ) -> Tuple[SimpleMetricInputInstance, ...]:
        """Update all simple-metric input instances with the corresponding fill_nulls_with value."""
        updated_instances: List[SimpleMetricInputInstance] = []
        for instance in instances:
            spec = instance.spec
            mapped_spec = self._null_fill_value_mapping.null_fill_value_spec(spec)
            if mapped_spec is not None:
                updated_instances.append(
                    SimpleMetricInputInstance(
                        associated_columns=instance.associated_columns,
                        spec=mapped_spec,
                        aggregation_state=instance.aggregation_state,
                        defined_from=instance.defined_from,
                    )
                )
            else:
                updated_instances.append(instance)

        return tuple(updated_instances)

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            simple_metric_input_instances=self._update_fill_nulls_with(instance_set.simple_metric_input_instances),
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )


class AliasAggregatedSimpleMetricInputs(InstanceSetTransform[InstanceSet]):
    """Returns a new instance set where all simple-metric inputs have been assigned an alias spec."""

    def __init__(self, alias_mapping: InstanceAliasMapping) -> None:
        """Initializer stores the input specs, which contain the aliases for each simple-metric input.

        Note this class only works if used in conjunction with an `AggregateSimpleMetricInputsNode` that has been generated
        by querying a single semantic model for a single set of aggregated simple-metric inputs. This is currently enforced
        by the structure of the `DataflowPlanBuilder`, which ensures each `AggregateSimpleMetricInputsNode` corresponds to
        a single semantic model set of simple-metric inputs for a single metric, and that these outputs will then be
        combined via joins.
        """
        self._alias_mapping = alias_mapping

    def _alias_simple_metric_input_instances(
        self, instances: Tuple[SimpleMetricInputInstance, ...]
    ) -> Tuple[SimpleMetricInputInstance, ...]:
        """Update all simple-metric input instances with aliases, if any are found in the input spec set."""
        aliased_instances: List[SimpleMetricInputInstance] = []
        for instance in instances:
            spec = instance.spec
            aliased_spec = self._alias_mapping.aliased_spec(spec)
            if aliased_spec is not None:
                aliased_instances.append(
                    SimpleMetricInputInstance(
                        associated_columns=instance.associated_columns,
                        spec=aliased_spec,
                        aggregation_state=instance.aggregation_state,
                        defined_from=instance.defined_from,
                    )
                )
            else:
                aliased_instances.append(instance)

        return tuple(aliased_instances)

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            simple_metric_input_instances=self._alias_simple_metric_input_instances(
                instance_set.simple_metric_input_instances
            ),
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )


class AddMetrics(InstanceSetTransform[InstanceSet]):
    """Adds the given metric instances to the instance set."""

    def __init__(self, metric_instances: List[MetricInstance]) -> None:  # noqa: D107
        self._metric_instances = metric_instances

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            simple_metric_input_instances=instance_set.simple_metric_input_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances + tuple(self._metric_instances),
            metadata_instances=instance_set.metadata_instances,
        )


class AddGroupByMetric(InstanceSetTransform[InstanceSet]):
    """Adds the given metric instances to the instance set."""

    def __init__(self, group_by_metric_instance: GroupByMetricInstance) -> None:  # noqa: D107
        self._group_by_metric_instance = group_by_metric_instance

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            simple_metric_input_instances=instance_set.simple_metric_input_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances + (self._group_by_metric_instance,),
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )


class RemoveSimpleMetricInputTransform(InstanceSetTransform[InstanceSet]):
    """Remove simple-metric inputs from the instance set."""

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            simple_metric_input_instances=(),
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )


class RemoveMetrics(InstanceSetTransform[InstanceSet]):
    """Remove metrics from the instance set."""

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            simple_metric_input_instances=instance_set.simple_metric_input_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=(),
            metadata_instances=instance_set.metadata_instances,
        )


class CreateSqlColumnReferencesForInstances(InstanceSetTransform[Tuple[SqlColumnReferenceExpression, ...]]):
    """Create select column expressions that will express all instances in the set.

    It assumes that the column names of the instances are represented by the supplied column association resolver and
    come from the given table alias.
    """

    def __init__(
        self,
        table_alias: str,
        column_resolver: ColumnAssociationResolver,
    ) -> None:
        """Initializer.

        Args:
            table_alias: the table alias to select columns from
            column_resolver: resolver to name columns.
        """
        self._table_alias = table_alias
        self._column_resolver = column_resolver

    def transform(self, instance_set: InstanceSet) -> Tuple[SqlColumnReferenceExpression, ...]:  # noqa: D102
        column_names = [
            self._column_resolver.resolve_spec(spec).column_name for spec in instance_set.spec_set.all_specs
        ]
        return tuple(
            SqlColumnReferenceExpression.create(
                SqlColumnReference(
                    table_alias=self._table_alias,
                    column_name=column_name,
                ),
            )
            for column_name in column_names
        )


class CreateSelectColumnForCombineOutputNode(InstanceSetTransform[SelectColumnSet]):
    """Create select column expressions for the instance for joining outputs.

    It assumes that the column names of the instances are represented by the supplied column association resolver and
    come from the given table alias.
    """

    def __init__(  # noqa: D107
        self,
        table_alias: str,
        column_resolver: ColumnAssociationResolver,
        metric_lookup: MetricLookup,
    ) -> None:
        self._table_alias = table_alias
        self._column_resolver = column_resolver
        self._metric_lookup = metric_lookup

    def _create_select_column(self, spec: InstanceSpec, fill_nulls_with: Optional[int] = None) -> SqlSelectColumn:
        """Creates the select column for the given spec and the fill value."""
        column_name = self._column_resolver.resolve_spec(spec).column_name
        column_reference_expression = SqlColumnReferenceExpression.create(
            col_ref=SqlColumnReference(
                table_alias=self._table_alias,
                column_name=column_name,
            )
        )
        select_expression: SqlExpressionNode = SqlFunctionExpression.build_expression_from_aggregation_type(
            aggregation_type=AggregationType.MAX, sql_column_expression=column_reference_expression
        )
        if fill_nulls_with is not None:
            select_expression = SqlAggregateFunctionExpression.create(
                sql_function=SqlFunction.COALESCE,
                sql_function_args=[
                    select_expression,
                    SqlStringExpression.create(str(fill_nulls_with)),
                ],
            )
        return SqlSelectColumn(
            expr=select_expression,
            column_alias=column_name,
        )

    def _create_select_columns_for_metrics(self, metric_instances: Tuple[MetricInstance, ...]) -> List[SqlSelectColumn]:
        select_columns: List[SqlSelectColumn] = []
        for metric_instance in metric_instances:
            metric_reference = MetricReference(element_name=metric_instance.defined_from.metric_name)
            metric = self._metric_lookup.get_metric(metric_reference)
            metric_type = metric.type
            if metric_type is MetricType.SIMPLE:
                fill_nulls_with = metric.type_params.fill_nulls_with
            elif (
                metric_type is MetricType.RATIO
                or metric_type is MetricType.DERIVED
                or metric_type is MetricType.CUMULATIVE
                or metric_type is MetricType.CONVERSION
            ):
                fill_nulls_with = None
            else:
                assert_values_exhausted(metric_type)
            select_columns.append(
                self._create_select_column(spec=metric_instance.spec, fill_nulls_with=fill_nulls_with)
            )
        return select_columns

    def _create_select_columns_for_simple_metric_inputs(
        self, instances: Tuple[SimpleMetricInputInstance, ...]
    ) -> List[SqlSelectColumn]:
        select_columns: List[SqlSelectColumn] = []
        for instance in instances:
            simple_metric_input_spec = instance.spec
            select_columns.append(
                self._create_select_column(
                    spec=simple_metric_input_spec, fill_nulls_with=simple_metric_input_spec.fill_nulls_with
                )
            )
        return select_columns

    def transform(self, instance_set: InstanceSet) -> SelectColumnSet:  # noqa: D102
        return SelectColumnSet.create(
            metric_columns=self._create_select_columns_for_metrics(instance_set.metric_instances),
            simple_metric_input_columns=self._create_select_columns_for_simple_metric_inputs(
                instance_set.simple_metric_input_instances
            ),
        )


# TODO: delete this class & all uses. It doesn't do anything.
class ChangeAssociatedColumns(InstanceSetTransform[InstanceSet]):
    """Change the columns associated with instances to the one specified by the resolver.

    This is useful for conveniently generating output instances for a node that serve as a "pass-through". The output
    instances can be a copy of the parent's instances, except that the column names need to be changed.

    e.g. the parent may have a data set:
        sql:
            SELECT
                is_lux AS is_lux_latext
            ...
        instance:
            DimensionInstance(column_name="is_lux", ...)

    but for the current node, we want a data set like:

        sql:
            SELECT
                is_lux_latest
                ...
            FROM (
                -- SQL from parent
                is_lux AS is_lux_latest
                ...
            )
            ...
        instance:
            DimensionInstance(column_name="is_lux_latest")
    """

    def __init__(self, column_association_resolver: ColumnAssociationResolver) -> None:  # noqa: D107
        self._column_association_resolver = column_association_resolver

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        output_simple_metric_input_instances = []
        for input_simple_metric_input_instance in instance_set.simple_metric_input_instances:
            output_simple_metric_input_instances.append(
                SimpleMetricInputInstance(
                    associated_columns=(
                        self._column_association_resolver.resolve_spec(input_simple_metric_input_instance.spec),
                    ),
                    spec=input_simple_metric_input_instance.spec,
                    defined_from=input_simple_metric_input_instance.defined_from,
                    aggregation_state=input_simple_metric_input_instance.aggregation_state,
                )
            )

        output_dimension_instances = []
        for input_dimension_instance in instance_set.dimension_instances:
            output_dimension_instances.append(
                DimensionInstance(
                    associated_columns=(self._column_association_resolver.resolve_spec(input_dimension_instance.spec),),
                    spec=input_dimension_instance.spec,
                    defined_from=input_dimension_instance.defined_from,
                )
            )

        output_time_dimension_instances = []
        for input_time_dimension_instance in instance_set.time_dimension_instances:
            output_time_dimension_instances.append(
                TimeDimensionInstance(
                    associated_columns=(
                        self._column_association_resolver.resolve_spec(input_time_dimension_instance.spec),
                    ),
                    spec=input_time_dimension_instance.spec,
                    defined_from=input_time_dimension_instance.defined_from,
                )
            )

        output_entity_instances = []
        for input_entity_instance in instance_set.entity_instances:
            output_entity_instances.append(
                EntityInstance(
                    associated_columns=(self._column_association_resolver.resolve_spec(input_entity_instance.spec),),
                    spec=input_entity_instance.spec,
                    defined_from=input_entity_instance.defined_from,
                )
            )

        output_metric_instances = []
        for input_metric_instance in instance_set.metric_instances:
            output_metric_instances.append(
                MetricInstance(
                    associated_columns=(self._column_association_resolver.resolve_spec(input_metric_instance.spec),),
                    spec=input_metric_instance.spec,
                    defined_from=input_metric_instance.defined_from,
                )
            )

        output_metadata_instances = []
        for input_metadata_instance in instance_set.metadata_instances:
            output_metadata_instances.append(
                MetadataInstance(
                    associated_columns=(self._column_association_resolver.resolve_spec(input_metadata_instance.spec),),
                    spec=input_metadata_instance.spec,
                )
            )

        output_group_by_metric_instances = []
        for input_group_by_metric_instance in instance_set.group_by_metric_instances:
            output_group_by_metric_instances.append(
                GroupByMetricInstance(
                    associated_columns=(
                        self._column_association_resolver.resolve_spec(input_group_by_metric_instance.spec),
                    ),
                    spec=input_group_by_metric_instance.spec,
                    defined_from=input_group_by_metric_instance.defined_from,
                )
            )
        return InstanceSet(
            simple_metric_input_instances=tuple(output_simple_metric_input_instances),
            dimension_instances=tuple(output_dimension_instances),
            time_dimension_instances=tuple(output_time_dimension_instances),
            entity_instances=tuple(output_entity_instances),
            group_by_metric_instances=tuple(output_group_by_metric_instances),
            metric_instances=tuple(output_metric_instances),
            metadata_instances=tuple(output_metadata_instances),
        )


class ConvertToMetadata(InstanceSetTransform[InstanceSet]):
    """Removes all instances from old instance set and replaces them with a set of metadata instances."""

    def __init__(self, metadata_instances: Sequence[MetadataInstance]) -> None:  # noqa: D107
        self._metadata_instances = metadata_instances

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            metadata_instances=tuple(self._metadata_instances),
        )


class AddMetadata(InstanceSetTransform[InstanceSet]):
    """Adds the given metric instances to the instance set."""

    def __init__(self, metadata_instances: Sequence[MetadataInstance]) -> None:  # noqa: D107
        self._metadata_instances = metadata_instances

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            simple_metric_input_instances=instance_set.simple_metric_input_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances + tuple(self._metadata_instances),
        )



from __future__ import annotations

import logging

from metricflow_semantics.experimental.orderd_enum import OrderedEnum

logger = logging.getLogger(__name__)


class LinkableElementType(OrderedEnum):
    """Enumeration of the possible types of linkable element we are encountering or expecting.

    Group-by items effectively map on to LinkableSpecs and queryable semantic manifest elements such
    as Metrics, Dimensions, and Entities. This provides the full set of types we might encounter, and is
    useful for ensuring that we are always getting the correct LinkableElement from a given part of the
    codebase - e.g., to ensure we are not accidentally getting an Entity when we expect a Dimension.
    """

    DIMENSION = "dimension"
    ENTITY = "entity"
    METRIC = "metric"
    TIME_DIMENSION = "time_dimension"



from __future__ import annotations

import dataclasses
import logging
import typing
from dataclasses import dataclass
from typing import Iterable

from metricflow_semantics.experimental.ordered_set import MutableOrderedSet

if typing.TYPE_CHECKING:
    from metricflow_semantics.experimental.semantic_graph.nodes.attribute_nodes import (
        CategoricalDimensionAttributeNode,
        KeyAttributeNode,
        TimeAttributeNode,
    )
    from metricflow_semantics.experimental.semantic_graph.nodes.entity_nodes import (
        ComplexMetricNode,
        ConfiguredEntityNode,
        JoinedModelNode,
        LocalModelNode,
        MetricTimeNode,
        SimpleMetricNode,
        TimeDimensionNode,
        TimeNode,
    )
    from metricflow_semantics.experimental.semantic_graph.sg_interfaces import SemanticGraphNode

logger = logging.getLogger(__name__)


@dataclass
class SemanticGraphNodeTypedCollection:
    """A mutable collection of the nodes in a semantic graph, grouped by type."""

    configured_entity_nodes: MutableOrderedSet[ConfiguredEntityNode] = dataclasses.field(
        default_factory=lambda: MutableOrderedSet()
    )
    joined_model_nodes: MutableOrderedSet[JoinedModelNode] = dataclasses.field(
        default_factory=lambda: MutableOrderedSet()
    )
    local_model_nodes: MutableOrderedSet[LocalModelNode] = dataclasses.field(
        default_factory=lambda: MutableOrderedSet()
    )
    time_dimension_nodes: MutableOrderedSet[TimeDimensionNode] = dataclasses.field(
        default_factory=lambda: MutableOrderedSet()
    )
    metric_time_nodes: MutableOrderedSet[MetricTimeNode] = dataclasses.field(
        default_factory=lambda: MutableOrderedSet()
    )
    time_nodes: MutableOrderedSet[TimeNode] = dataclasses.field(default_factory=lambda: MutableOrderedSet())
    simple_metric_nodes: MutableOrderedSet[SimpleMetricNode] = dataclasses.field(
        default_factory=lambda: MutableOrderedSet()
    )
    complex_metric_nodes: MutableOrderedSet[ComplexMetricNode] = dataclasses.field(
        default_factory=lambda: MutableOrderedSet()
    )

    time_attribute_nodes: MutableOrderedSet[TimeAttributeNode] = dataclasses.field(
        default_factory=lambda: MutableOrderedSet()
    )
    key_attribute_nodes: MutableOrderedSet[KeyAttributeNode] = dataclasses.field(
        default_factory=lambda: MutableOrderedSet()
    )
    categorical_dimension_attribute_nodes: MutableOrderedSet[CategoricalDimensionAttributeNode] = dataclasses.field(
        default_factory=lambda: MutableOrderedSet()
    )

    @staticmethod
    def create(nodes: Iterable[SemanticGraphNode]) -> SemanticGraphNodeTypedCollection:  # noqa: D102
        node_collection = SemanticGraphNodeTypedCollection()
        for node in nodes:
            node.add_to_typed_collection(node_collection)
        return node_collection



from __future__ import annotations

import json
import pathlib
from pathlib import Path
from typing import Dict, Optional

from dbt_semantic_interfaces.implementations.metric import PydanticMetric
from dbt_semantic_interfaces.implementations.node_relation import PydanticNodeRelation
from dbt_semantic_interfaces.implementations.project_configuration import PydanticProjectConfiguration
from dbt_semantic_interfaces.implementations.saved_query import PydanticSavedQuery
from dbt_semantic_interfaces.implementations.semantic_manifest import PydanticSemanticManifest
from dbt_semantic_interfaces.implementations.semantic_model import PydanticSemanticModel
from dbt_semantic_interfaces.implementations.time_spine import PydanticTimeSpine, PydanticTimeSpinePrimaryColumn
from dbt_semantic_interfaces.parsing.dir_to_model import (
    parse_directory_of_yaml_files_to_semantic_manifest,
)
from dbt_semantic_interfaces.type_enums import TimeGranularity
from dbt_semantic_interfaces.validations.semantic_manifest_validator import SemanticManifestValidator

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat


def mf_load_manifest_from_yaml_directory(
    yaml_file_directory: pathlib.Path,
    template_mapping: Optional[Dict[str, str]] = None,
) -> PydanticSemanticManifest:
    """Reads the manifest YAMLs from the standard location, applies transformations, runs validations."""
    try:
        build_result = parse_directory_of_yaml_files_to_semantic_manifest(
            str(yaml_file_directory), template_mapping=template_mapping
        )
        validator = SemanticManifestValidator[PydanticSemanticManifest]()
        validator.checked_validations(build_result.semantic_manifest)
        return build_result.semantic_manifest
    except Exception as e:
        raise RuntimeError(
            LazyFormat("Error while loading semantic manifest", yaml_file_directory=yaml_file_directory)
        ) from e


def mf_load_manifest_from_json_file(
    json_file_path: Path, project_configuration: Optional[PydanticProjectConfiguration] = None
) -> PydanticSemanticManifest:
    """Load a manifest from a file containing the JSON-serialized form of a `PydanticSemanticManifest`.

    If the project configuration is not provided, a dummy one with an `HOUR` time spine will be used.
    """
    try:
        with open(json_file_path) as fp:
            manifest_json = json.load(fp)

        semantic_models = [
            PydanticSemanticModel.parse_obj(semantic_model_json)
            for semantic_model_json in manifest_json["semantic_models"]
        ]
        metrics = [PydanticMetric.parse_obj(metric_json) for metric_json in manifest_json["metrics"]]
        saved_queries = [
            PydanticSavedQuery.parse_obj(saved_query_json) for saved_query_json in manifest_json["saved_queries"]
        ]

        if project_configuration is None:
            node_relation = PydanticNodeRelation(
                schema_name="dummy_schema",
                relation_name="dummy_relation",
                database="dummy_database",
                alias="dummy_alias",
            )
            project_configuration = PydanticProjectConfiguration(
                time_spines=[
                    PydanticTimeSpine(
                        node_relation=node_relation,
                        primary_column=PydanticTimeSpinePrimaryColumn(
                            name="date_hour",
                            time_granularity=TimeGranularity.HOUR,
                        ),
                    )
                ],
            )

        return PydanticSemanticManifest(
            semantic_models=semantic_models,
            metrics=metrics,
            saved_queries=saved_queries,
            project_configuration=project_configuration,
        )

    except Exception as e:
        raise RuntimeError(LazyFormat("Error while loading semantic manifest", json_file_path=json_file_path)) from e



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple

from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput
from metricflow_semantics.query.resolver_inputs.query_resolver_inputs import (
    ResolverInputForMetric,
)


@dataclass(frozen=True)
class InvalidApplyGroupByIssue(MetricFlowQueryResolutionIssue):
    """Describes an issue with the query where the apply group by param is invalid."""

    apply_group_by: bool
    metric_inputs: Tuple[ResolverInputForMetric, ...]

    @staticmethod
    def from_parameters(  # noqa: D102
        apply_group_by: bool,
        metric_inputs: Tuple[ResolverInputForMetric, ...],
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> InvalidApplyGroupByIssue:
        return InvalidApplyGroupByIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            query_resolution_path=query_resolution_path,
            apply_group_by=apply_group_by,
            metric_inputs=metric_inputs,
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        return "`apply_group_by` cannot be False for a query with metrics."

    @override
    def with_path_prefix(self, path_prefix_node: MetricFlowQueryResolutionPath) -> InvalidApplyGroupByIssue:
        return InvalidApplyGroupByIssue(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix_node) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix_node),
            apply_group_by=self.apply_group_by,
            metric_inputs=self.metric_inputs,
        )



from __future__ import annotations

import itertools
import logging
import typing
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Iterable, Optional, Sequence, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.references import EntityReference, SemanticModelReference
from dbt_semantic_interfaces.type_enums import DatePart
from typing_extensions import Self

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.model.semantic_model_derivation import SemanticModelDerivation
from metricflow_semantics.model.semantics.element_filter import GroupByItemSetFilter
from metricflow_semantics.model.semantics.linkable_element import LinkableElementType
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.group_by_metric_spec import GroupByMetricSpec
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

if typing.TYPE_CHECKING:
    from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe import IndexedDunderName
    from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_descriptor import (
        DunderNameDescriptor,
    )

logger = logging.getLogger(__name__)


class BaseGroupByItemSet(SemanticModelDerivation, ABC):
    """ABC for a set that represents possible group-by items."""

    @abstractmethod
    def filter(
        self,
        element_filter: GroupByItemSetFilter,
    ) -> Self:
        """Filter elements in the set.

        First, only elements with at least one property in the "any_properties_allowlist" set are retained. Then, any elements with
        a property in "any_properties_denylist" set are removed.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_empty(self) -> bool:
        """Whether this maps to an empty set of specs.

        If `.specs` is created lazily, there are cases where the caller only needs to know that there are matching
        specs.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def specs(self) -> Sequence[LinkableInstanceSpec]:
        """Converts the items represented by this to the corresponding spec objects."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def annotated_specs(self) -> Sequence[AnnotatedSpec]:
        """Return the annotated specs represented by this."""
        raise NotImplementedError

    @abstractmethod
    def filter_by_spec_patterns(self, spec_patterns: Sequence[SpecPattern]) -> Self:
        """Filter the elements in the set by the given spec patters.

        Returns a new set consisting of the elements in the `LinkableElementSet` that have a corresponding spec that
        match all the given spec patterns.
        """
        raise NotImplementedError()

    @abstractmethod
    def intersection(self, *others: Self) -> Self:
        """Intersect this set with the other set to determine possible group-by items."""
        raise NotImplementedError

    @abstractmethod
    def union(self, *others: Self) -> Self:
        """Intersect this set with the other set to determine possible group-by items."""
        raise NotImplementedError


@fast_frozen_dataclass()
class AnnotatedSpec(SerializableDataclass):
    """Groups a spec with more context about the spec.

    * This is intended to consolidate the `LinkableElement` classes during migration.
    * This class will be updated / renamed.
    """

    element_type: LinkableElementType
    element_name: str
    entity_link_names: Tuple[str, ...]
    time_grain: Optional[ExpandedTimeGranularity]
    date_part: Optional[DatePart]
    metric_subquery_entity_link_names: Tuple[str, ...]

    element_properties: Tuple[GroupByItemProperty, ...]
    # The semantic model(s) where the element (e.g. the categorical dimension) was defined.
    # There can be multiple models if it's a metric / derived metric that references multiple simple-metric inputs, and the join
    # path from the simple-metric input to the dimension is different.
    origin_semantic_model_names: Tuple[str, ...]
    derived_from_semantic_model_names: Tuple[str, ...]

    @staticmethod
    def create(  # noqa: D102
        element_type: LinkableElementType,
        element_name: str,
        properties: Iterable[GroupByItemProperty],
        origin_model_ids: Iterable[SemanticModelId],
        derived_from_semantic_models: Iterable[SemanticModelReference],
        entity_links: Sequence[EntityReference],
        metric_subquery_entity_links: Optional[Sequence[EntityReference]],
        time_grain: Optional[ExpandedTimeGranularity],
        date_part: Optional[DatePart],
    ) -> AnnotatedSpec:
        entity_link_names = tuple(entity_reference.element_name for entity_reference in entity_links)
        element_properties = tuple(FrozenOrderedSet(properties))
        origin_model_names = tuple(FrozenOrderedSet(model_id.model_name for model_id in origin_model_ids))
        derived_from_semantic_model_names = tuple(
            FrozenOrderedSet(model_reference.semantic_model_name for model_reference in derived_from_semantic_models)
        )
        metric_subquery_entity_link_names = (
            tuple(entity_reference.element_name for entity_reference in metric_subquery_entity_links)
            if metric_subquery_entity_links is not None
            else ()
        )

        return AnnotatedSpec(
            element_type=element_type,
            element_name=element_name,
            entity_link_names=entity_link_names,
            element_properties=element_properties,
            time_grain=time_grain,
            date_part=date_part,
            metric_subquery_entity_link_names=metric_subquery_entity_link_names,
            origin_semantic_model_names=origin_model_names,
            derived_from_semantic_model_names=derived_from_semantic_model_names,
        )

    @staticmethod
    def create_from_indexed_dunder_name(  # noqa: D102
        indexed_dunder_name: IndexedDunderName, descriptor: DunderNameDescriptor
    ) -> Sequence[AnnotatedSpec]:
        items: list[AnnotatedSpec] = []

        element_type = descriptor.element_type

        if (
            element_type is LinkableElementType.DIMENSION
            or element_type is LinkableElementType.ENTITY
            or element_type is LinkableElementType.METRIC
        ):
            element_name = indexed_dunder_name[-1]
            entity_links = tuple(EntityReference(name_element) for name_element in indexed_dunder_name[:-1])
        elif element_type is LinkableElementType.TIME_DIMENSION:
            element_name = indexed_dunder_name[-2]
            entity_links = tuple(EntityReference(name_element) for name_element in indexed_dunder_name[:-2])
        else:
            assert_values_exhausted(element_type)

        if (
            element_type is LinkableElementType.DIMENSION
            or element_type is LinkableElementType.ENTITY
            or element_type is LinkableElementType.TIME_DIMENSION
        ):
            items.append(
                AnnotatedSpec.create(
                    element_type=descriptor.element_type,
                    element_name=element_name,
                    properties=descriptor.element_properties,
                    origin_model_ids=descriptor.origin_model_ids,
                    derived_from_semantic_models=(
                        model_id.semantic_model_reference for model_id in descriptor.derived_from_model_ids
                    ),
                    entity_links=entity_links,
                    metric_subquery_entity_links=None,
                    time_grain=descriptor.time_grain,
                    date_part=descriptor.date_part,
                )
            )
        elif element_type is LinkableElementType.METRIC:
            for entity_key_query in descriptor.entity_key_queries_for_group_by_metric:
                items.append(
                    AnnotatedSpec.create(
                        element_type=descriptor.element_type,
                        element_name=element_name,
                        properties=descriptor.element_properties,
                        origin_model_ids=descriptor.origin_model_ids,
                        derived_from_semantic_models=FrozenOrderedSet(
                            itertools.chain(
                                (model_id.semantic_model_reference for model_id in descriptor.derived_from_model_ids),
                                (
                                    model_id.semantic_model_reference
                                    for model_id in entity_key_query.derived_from_model_ids
                                ),
                            )
                        ),
                        entity_links=entity_links,
                        metric_subquery_entity_links=tuple(
                            EntityReference(name_element) for name_element in entity_key_query.entity_key_query
                        ),
                        time_grain=None,
                        date_part=None,
                    )
                )
        else:
            assert_values_exhausted(element_type)

        return items

    @cached_property
    def spec(self) -> LinkableInstanceSpec:
        """Return the type-specific spec object.

        This should be renamed to avoid confusion.
        """
        element_type = self.element_type
        element_name = self.element_name
        entity_links = tuple(EntityReference(entity_link_name) for entity_link_name in self.entity_link_names)
        if element_type is LinkableElementType.METRIC:
            metric_subquery_entity_link_names = self.metric_subquery_entity_link_names
            assert metric_subquery_entity_link_names is not None
            metric_subquery_entity_links = tuple(
                EntityReference(entity_link_name) for entity_link_name in metric_subquery_entity_link_names
            )
            return GroupByMetricSpec(
                element_name=element_name,
                entity_links=entity_links,
                metric_subquery_entity_links=metric_subquery_entity_links,
            )
        elif element_type is LinkableElementType.TIME_DIMENSION:
            return TimeDimensionSpec(
                element_name=element_name,
                entity_links=entity_links,
                time_granularity=self.time_grain,
                date_part=self.date_part,
            )
        elif element_type is LinkableElementType.DIMENSION:
            return DimensionSpec(
                element_name=element_name,
                entity_links=entity_links,
            )
        elif element_type is LinkableElementType.ENTITY:
            return EntitySpec(
                element_name=self.element_name,
                entity_links=entity_links,
            )
        elif element_type is LinkableElementType.TIME_DIMENSION:
            return TimeDimensionSpec(
                element_name=element_name,
                entity_links=entity_links,
                time_granularity=self.time_grain,
                date_part=self.date_part,
            )
        else:
            assert_values_exhausted(element_type)

    @cached_property
    def property_set(self) -> FrozenOrderedSet[GroupByItemProperty]:  # noqa: D102
        return FrozenOrderedSet(self.element_properties)

    @cached_property
    def origin_model_ids(self) -> FrozenOrderedSet[SemanticModelId]:  # noqa: D102
        return FrozenOrderedSet(
            SemanticModelId.get_instance(model_name) for model_name in self.origin_semantic_model_names
        )

    @cached_property
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:  # noqa: D102
        return tuple(SemanticModelReference(model_name) for model_name in self.derived_from_semantic_model_names)

    def merge(self, other: AnnotatedSpec) -> AnnotatedSpec:  # noqa: D102
        if (
            self.element_type != other.element_type
            or self.element_name != other.element_name
            or self.time_grain != other.time_grain
            or self.date_part != other.date_part
            or self.metric_subquery_entity_link_names != other.metric_subquery_entity_link_names
        ):
            raise RuntimeError(
                LazyFormat(
                    "Unable to merge annotated specs due to incompatible fields. This indicates an error in"
                    " the set of annotated specs being merged.",
                    self=self,
                    other=other,
                )
            )

        return AnnotatedSpec(
            element_type=self.element_type,
            element_name=self.element_name,
            entity_link_names=self.entity_link_names,
            time_grain=self.time_grain,
            date_part=self.date_part,
            metric_subquery_entity_link_names=self.metric_subquery_entity_link_names,
            element_properties=tuple(self.property_set.union(other.property_set)),
            origin_semantic_model_names=tuple(
                FrozenOrderedSet(self.origin_semantic_model_names + other.origin_semantic_model_names)
            ),
            derived_from_semantic_model_names=tuple(
                FrozenOrderedSet(self.derived_from_semantic_model_names + other.derived_from_semantic_model_names)
            ),
        )

    @cached_property
    def origin_semantic_model_references(self) -> AnyLengthTuple[SemanticModelReference]:
        """Returns the models where the element is defined.

        There can be multiple as a derived metric may have a group-by item that can be reached via multiple join paths.
        """
        return tuple(model_id.semantic_model_reference for model_id in self.origin_model_ids)



from __future__ import annotations

from typing import List

from dbt_semantic_interfaces.implementations.node_relation import PydanticNodeRelation
from dbt_semantic_interfaces.implementations.project_configuration import PydanticProjectConfiguration
from dbt_semantic_interfaces.implementations.semantic_manifest import PydanticSemanticManifest
from dbt_semantic_interfaces.implementations.semantic_model import PydanticSemanticModel
from dbt_semantic_interfaces.implementations.time_spine import PydanticTimeSpine, PydanticTimeSpinePrimaryColumn
from dbt_semantic_interfaces.type_enums import TimeGranularity

from metricflow_semantics.test_helpers.synthetic_manifest.categorical_dimension_generator import (
    CategoricalDimensionGenerator,
)
from metricflow_semantics.test_helpers.synthetic_manifest.dimension_semantic_model_generator import (
    DimensionSemanticModelGenerator,
)
from metricflow_semantics.test_helpers.synthetic_manifest.metric_generator import MetricGenerator
from metricflow_semantics.test_helpers.synthetic_manifest.saved_query_generator import SavedQueryGenerator
from metricflow_semantics.test_helpers.synthetic_manifest.simple_metric_semantic_model_generator import (
    SimpleMetricSemanticModelGenerator,
)
from metricflow_semantics.test_helpers.synthetic_manifest.synthetic_manifest_parameter_set import (
    SyntheticManifestParameterSet,
)


class SyntheticManifestGenerator:
    """Generates a synthetic semantic manifest that can be used for performance profiling."""

    def __init__(self, parameter_set: SyntheticManifestParameterSet) -> None:  # noqa: D107
        self._parameter_set = parameter_set
        self._categorical_dimension_generator = CategoricalDimensionGenerator(parameter_set)
        self._simple_metric_semantic_model_generator = SimpleMetricSemanticModelGenerator(
            parameter_set=parameter_set,
        )
        self._dimension_semantic_model_generator = DimensionSemanticModelGenerator(
            parameter_set=parameter_set,
            categorical_dimension_generator=self._categorical_dimension_generator,
        )
        self._metric_generator = MetricGenerator(
            parameter_set=parameter_set,
            semantic_model_generator=self._simple_metric_semantic_model_generator,
        )
        self._saved_query_generator = SavedQueryGenerator(
            parameter_set=parameter_set,
            metric_generator=self._metric_generator,
            categorical_dimension_generator=self._categorical_dimension_generator,
        )

    def generate_manifest(self) -> PydanticSemanticManifest:
        """Generate a manifest using the given parameters."""
        semantic_models: List[PydanticSemanticModel] = []

        semantic_models.extend(self._simple_metric_semantic_model_generator.generate_semantic_models())
        semantic_models.extend(self._dimension_semantic_model_generator.generate_semantic_models())

        return PydanticSemanticManifest(
            semantic_models=semantic_models,
            metrics=list(self._metric_generator.generate_metrics()),
            project_configuration=PydanticProjectConfiguration(
                time_spines=[
                    PydanticTimeSpine(
                        node_relation=PydanticNodeRelation(
                            alias="time_spine_source_table",
                            schema_name="demo",
                        ),
                        primary_column=PydanticTimeSpinePrimaryColumn(
                            name="ds",
                            time_granularity=TimeGranularity.DAY,
                        ),
                    )
                ]
            ),
            saved_queries=self._saved_query_generator.generate_saved_queries(),
        )



from __future__ import annotations

import logging
from collections.abc import Set
from typing import Optional

from typing_extensions import override

from metricflow_semantics.experimental.cache.mf_cache import ResultCache
from metricflow_semantics.experimental.metricflow_exception import MetricFlowInternalError
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.mf_graph.path_finding.pathfinder import MetricFlowPathfinder
from metricflow_semantics.experimental.mf_graph.path_finding.traversal_profile_differ import TraversalProfileDiffer
from metricflow_semantics.experimental.ordered_set import OrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe import AttributeRecipe
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.recipe_writer_path import (
    AttributeRecipeWriterPath,
)
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.recipe_writer_weight import (
    AttributeRecipeWriterWeightFunction,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraph,
    SemanticGraphEdge,
    SemanticGraphNode,
)
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_descriptor import DunderNameDescriptor
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_trie import (
    DunderNameTrie,
    MutableDunderNameTrie,
)
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_trie_resolver import (
    DunderNameTrieResolver,
    TrieCacheKey,
    TrieResolutionResult,
)
from metricflow_semantics.helpers.performance_helpers import ExecutionTimer
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.model.semantics.element_filter import GroupByItemSetFilter
from metricflow_semantics.model.semantics.semantic_model_join_evaluator import MAX_JOIN_HOPS

logger = logging.getLogger(__name__)


class SimpleTrieResolver(DunderNameTrieResolver):
    """Resolves the dunder-name trie that represents the "simple" group-by items available for metrics.

    The set of simple group-by items does not include group-by metrics. Those are handled in a separate resolver.
    """

    def __init__(
        self,
        semantic_graph: SemanticGraph,
        path_finder: MetricFlowPathfinder[SemanticGraphNode, SemanticGraphEdge, AttributeRecipeWriterPath],
        max_path_model_count: Optional[int] = None,
    ) -> None:
        """Initializer.

        Args:
            semantic_graph: The semantic graph that should be traversed to resolve the available group-by items.
            path_finder: The pathfinder to use for traversal.
            max_path_model_count: To handle special cases, the number of semantic models in a path can be limited by
            setting this argument.
        """
        super().__init__(semantic_graph=semantic_graph, path_finder=path_finder)
        self._verbose_debug_logs = False
        self._result_cache: ResultCache[TrieCacheKey, DunderNameTrie] = ResultCache()
        self._max_path_model_count = max_path_model_count

    @override
    def resolve_trie(
        self, source_nodes: OrderedSet[SemanticGraphNode], element_filter: Optional[GroupByItemSetFilter]
    ) -> TrieResolutionResult:
        execution_timer = ExecutionTimer()
        traversal_stat_differ = TraversalProfileDiffer(self._path_finder)
        with execution_timer, traversal_stat_differ:
            result_trie = self._resolve_trie_for_source_nodes(source_nodes=source_nodes, element_filter=element_filter)

        return TrieResolutionResult(
            duration=execution_timer.total_duration,
            traversal_profile=traversal_stat_differ.profile_delta,
            dunder_name_trie=result_trie,
        )

    def _resolve_trie_for_source_nodes(
        self,
        source_nodes: OrderedSet[SemanticGraphNode],
        element_filter: Optional[GroupByItemSetFilter],
    ) -> DunderNameTrie:
        """Resolve the available group-by items for the given source nodes.

        The source nodes can be metric nodes, measure nodes, or local-model nodes. Similar to the query interface,
        the set of group-by items that are available for the source nodes is the intersection of the items that are
        available for each node.
        """
        # Find the set simple-metric inputs / local-model nodes that the given source nodes depend on. Generating the result for
        # the set of given source nodes requires intersecting the result produced from each node.
        find_descendants_result = self._path_finder.find_descendants(
            graph=self._semantic_graph,
            source_nodes=source_nodes,
            target_nodes=self._semantic_graph.nodes_with_labels(
                self._simple_metric_label,
                self._local_model_label,
                self._metric_time_label,
            ),
            node_allow_set=self._semantic_graph.nodes_with_labels(
                self._metric_label,
                self._local_model_label,
                self._metric_time_label,
            ),
            deny_labels={self._deny_visible_attributes_label},
        )

        # Match the behavior of a bug in the existing resolver in some cases with date part. This should be removed
        # after migration.
        if self._should_exclude_date_part(
            source_nodes=source_nodes, collected_labels=find_descendants_result.labels_collected_during_traversal
        ):
            if element_filter is None:
                element_filter = GroupByItemSetFilter.create()
            element_filter = element_filter.copy(
                any_properties_denylist=element_filter.any_properties_denylist.union((GroupByItemProperty.DATE_PART,))
            )

        result_intersection_source_nodes = tuple(find_descendants_result.reachable_target_nodes)
        if self._verbose_debug_logs:
            logger.debug(
                LazyFormat(
                    "Found result-intersection nodes.",
                    find_descendants_result=find_descendants_result,
                )
            )

        if len(result_intersection_source_nodes) == 0:
            raise MetricFlowInternalError(
                LazyFormat("No applicable descendant nodes were found for intersection.", source_nodes=source_nodes)
            )

        result_trie = MutableDunderNameTrie.intersection_merge_common(
            tuple(
                self._resolve_trie_from_node(intersection_source_node, element_filter)
                for intersection_source_node in result_intersection_source_nodes
            )
        )

        if self._verbose_debug_logs:
            logger.debug(
                LazyFormat(
                    "Resolved intersection trie.",
                    result_trie=result_trie,
                )
            )
        return result_trie

    def _resolve_trie_from_node(
        self,
        source_node: SemanticGraphNode,
        element_filter: Optional[GroupByItemSetFilter],
    ) -> DunderNameTrie:
        source_node_labels = source_node.labels
        if self._local_model_label in source_node_labels or self._metric_time_label in source_node_labels:
            return self._resolve_trie_from_initial_path(
                initial_path=AttributeRecipeWriterPath.create(source_node),
                element_filter=element_filter,
            )

        # A simple-metric node has 2 successors, a local-model node that represents where the simple metric is defined,
        # and a metric-time node. Since many simple-metric nodes point to the same local-model node, generate and
        # cache results separately for each successor edge.
        elif self._simple_metric_label in source_node_labels:
            successors = self._semantic_graph.successors(source_node)
            local_model_edge: Optional[SemanticGraphEdge] = None
            metric_time_edge: Optional[SemanticGraphEdge] = None
            edges_from_source_node = self._semantic_graph.edges_with_tail_node(source_node)
            for edge in edges_from_source_node:
                if self._local_model_label in edge.head_node.labels:
                    local_model_edge = edge
                elif self._metric_time_label in edge.head_node.labels:
                    metric_time_edge = edge

            if local_model_edge is None or metric_time_edge is None or len(successors) != 2:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "A measure node should have exactly 2 successors: a local-model node and a metric-time node"
                        "\nbut it does not.",
                        source_node=source_node,
                        source_node_labels=source_node_labels,
                        local_model_edge=local_model_edge,
                        metric_time_edge=metric_time_edge,
                        edges_from_source_node=edges_from_source_node,
                    )
                )
            result_from_local_model_node = self._resolve_trie_from_initial_path(
                initial_path=AttributeRecipeWriterPath.create(local_model_edge.head_node),
                element_filter=element_filter,
            )

            result_from_metric_time = self._resolve_trie_from_initial_path(
                initial_path=AttributeRecipeWriterPath.create_from_edge(metric_time_edge, 0),
                element_filter=element_filter,
            )
            union_result = MutableDunderNameTrie.union_exclude_common(
                (result_from_local_model_node, result_from_metric_time)
            )

            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Resolved trie union",
                        local_model=result_from_local_model_node.dunder_names(),
                        metric_time=result_from_metric_time.dunder_names(),
                        union_result=union_result.dunder_names(),
                    )
                )

            return MutableDunderNameTrie.union_exclude_common((result_from_local_model_node, result_from_metric_time))
        else:
            raise MetricFlowInternalError(
                LazyFormat(
                    "The given node is not a supported source node for resolving attribute names",
                    source_node=source_node,
                )
            )

    def _resolve_trie_from_initial_path(
        self,
        initial_path: AttributeRecipeWriterPath,
        element_filter: Optional[GroupByItemSetFilter],
    ) -> DunderNameTrie:
        """Resolve the available group-by items using the given initial path."""
        cache_key = TrieCacheKey(key_nodes=tuple(initial_path.nodes), element_filter=element_filter)
        result = self._result_cache.get(cache_key)
        if result:
            return result.value

        target_nodes = self._semantic_graph.nodes_with_labels(self._group_by_attribute_label)
        if self._verbose_debug_logs:
            logger.debug(
                LazyFormat(
                    "Resolving names",
                    initial_path=initial_path,
                    element_filter=element_filter,
                )
            )

        result_trie = MutableDunderNameTrie()

        recipes: list[AttributeRecipe] = []
        for found_path in self._path_finder.find_paths_dfs(
            graph=self._semantic_graph,
            initial_path=initial_path,
            target_nodes=target_nodes,
            weight_function=AttributeRecipeWriterWeightFunction(
                element_filter, max_path_model_count=self._max_path_model_count
            ),
            max_path_weight=MAX_JOIN_HOPS,
            node_allow_set=None,
            node_deny_set=None,
        ):
            recipe = found_path.latest_recipe
            if self._verbose_debug_logs:
                logger.debug(LazyFormat("Found path to target node", path_nodes=found_path.nodes, recipe=recipe))
            if recipe is None:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "A path from a source node to a target node does not have a recipe. This indicates an error in "
                        "traversal."
                    )
                )
            recipes.append(recipe)

        result_trie.add_name_items(
            [
                (
                    recipe.indexed_dunder_name,
                    DunderNameDescriptor(
                        element_type=recipe.element_type,
                        time_grain=recipe.recipe_time_grain,
                        date_part=recipe.recipe_date_part,
                        element_properties=tuple(recipe.resolve_complete_properties()),
                        origin_model_ids=(recipe.joined_model_ids[-1],)
                        if recipe.joined_model_ids
                        else self._virtual_semantic_model_ids,
                        derived_from_model_ids=recipe.joined_model_ids
                        if recipe.joined_model_ids
                        else self._virtual_semantic_model_ids,
                        entity_key_queries_for_group_by_metric=(),
                    ),
                )
                for recipe in recipes
                if recipe.element_type is not None
            ]
        )

        return self._result_cache.set_and_get(cache_key, result_trie)

    def _should_exclude_date_part(
        self, source_nodes: Set[SemanticGraphNode], collected_labels: OrderedSet[MetricFlowGraphLabel]
    ) -> bool:
        """To handle a bug, check if we should exclude date part from the query.

        This reproduces the logic for what seems like a bug in the current resolver. The date part is only not allowed
        if there's a cumulative metric in the query. A query for a derived metric that is based on a cumulative metric
        still can be queried by date part.
        """
        if self._deny_date_part_label not in collected_labels:
            return False

        for source_node in source_nodes:
            for successor_edge in self._semantic_graph.edges_with_tail_node(source_node):
                if self._cumulative_metric_label in successor_edge.labels:
                    return True
        return False



from __future__ import annotations

import logging
from typing import Optional

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.singleton import Singleton

logger = logging.getLogger(__name__)


@fast_frozen_dataclass(order=False)
class GroupByAttributeLabel(MetricFlowGraphLabel, Singleton):
    """Label for any attribute node that can be used in the group-by argument of an MF query."""

    @classmethod
    def get_instance(cls) -> GroupByAttributeLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class ConfiguredEntityLabel(MetricFlowGraphLabel, Singleton):
    """Label for nodes that correspond to entities configured in a semantic model."""

    @classmethod
    def get_instance(cls) -> ConfiguredEntityLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class TimeDimensionLabel(MetricFlowGraphLabel, Singleton):
    """Label for time dimension nodes."""

    @classmethod
    def get_instance(cls) -> TimeDimensionLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class TimeClusterLabel(MetricFlowGraphLabel, Singleton):
    """Label for nodes that should be clustered together in the `time` section when visualizing the graph."""

    @classmethod
    def get_instance(cls) -> TimeClusterLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class MetricTimeLabel(MetricFlowGraphLabel, Singleton):
    """Label for the node that represents metric-time."""

    @classmethod
    def get_instance(cls) -> MetricTimeLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class MetricLabel(MetricFlowGraphLabel, Singleton):
    """Label for the node that corresponds to a configured metric in the semantic manifest.

    `metric_name = None` is a label applied to all metric nodes.
    """

    metric_name: Optional[str]

    @classmethod
    def get_instance(cls, metric_name: Optional[str] = None) -> MetricLabel:  # noqa: D102
        return cls._get_instance(metric_name=metric_name)


@fast_frozen_dataclass(order=False)
class SimpleMetricLabel(MetricFlowGraphLabel, Singleton):
    """Label for nodes that represent a simple metric."""

    @classmethod
    def get_instance(cls) -> SimpleMetricLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class ComplexMetricLabel(MetricFlowGraphLabel, Singleton):
    """Label for nodes that represent a complex metric."""

    @classmethod
    def get_instance(cls) -> ComplexMetricLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class JoinedModelLabel(MetricFlowGraphLabel, Singleton):
    """Label for nodes that represent a joined semantic model.

    See `JoinedModelNode`.
    """

    @classmethod
    def get_instance(cls) -> JoinedModelLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class LocalModelLabel(MetricFlowGraphLabel, Singleton):
    """Label for nodes that represent a local semantic model.

    See `LocalModelNode`.
    """

    @classmethod
    def get_instance(cls) -> LocalModelLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class KeyAttributeLabel(MetricFlowGraphLabel, Singleton):
    """Label for nodes that correspond to the entity-key attribute nodes."""

    @classmethod
    def get_instance(cls) -> KeyAttributeLabel:  # noqa: D102
        return cls._get_instance()



from __future__ import annotations

import logging

import termcolor

logger = logging.getLogger(__name__)


def mf_colored_link_text(uri: str) -> str:
    """Generates a string with color codes that looks like a link for logging on the terminal.

    Using `termcolor` to handle the cases where terminal color has been disabled, and we don't want to be printing
    color codes (e.g. `NO_COLOR` https://no-color.org/).
    """
    return termcolor.colored(uri, "blue", attrs=["bold"])



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Optional

from typing_extensions import override

from metricflow_semantics.experimental.mf_graph.path_finding.graph_path import MutableGraphPath
from metricflow_semantics.experimental.mf_graph.path_finding.pathfinder import MetricFlowPathfinder
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe import (
    AttributeRecipe,
)
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStep,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import SemanticGraphEdge, SemanticGraphNode
from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable
from metricflow_semantics.mf_logging.pretty_formatter import PrettyFormatContext

logger = logging.getLogger(__name__)

_EMPTY_RECIPE = AttributeRecipe()


@dataclass
class AttributeRecipeWriterPath(MutableGraphPath[SemanticGraphNode, SemanticGraphEdge], MetricFlowPrettyFormattable):
    """An implementation of a path in the semantic graph that writes recipes.

    The nodes and edges in the semantic graph are annotated with recipe steps that describe the computation of
    attributes. This path takes those steps and merges them into a single recipe as nodes are added to the path.

    In DFS traversal, nodes are added and popped at the end. To support this operation, recipe versions are
    stored in a list so that consistent state can be maintained.
    """

    # Every time a node / edge is added, the updated recipe is added to this list.
    _recipe_versions: list[AttributeRecipe]

    @staticmethod
    def create(start_node: Optional[SemanticGraphNode] = None) -> AttributeRecipeWriterPath:  # noqa: D102
        path = AttributeRecipeWriterPath(
            _nodes=[],
            _edges=[],
            _weight_addition_order=[],
            _current_weight=0,
            _current_node_set=set(),
            _node_set_addition_order=[],
            _recipe_versions=[_EMPTY_RECIPE],
        )
        if start_node:
            path._append_node(start_node)
            path._append_step(start_node.recipe_step_to_append)
        return path

    @staticmethod
    def create_from_edge(start_edge: SemanticGraphEdge, weight: int) -> AttributeRecipeWriterPath:  # noqa: D102
        path = AttributeRecipeWriterPath(
            _nodes=[],
            _edges=[],
            _weight_addition_order=[],
            _current_weight=0,
            _current_node_set=set(),
            _node_set_addition_order=[],
            _recipe_versions=[_EMPTY_RECIPE],
        )
        path.append_edge(start_edge, weight)
        return path

    def _append_step(self, recipe_step: AttributeRecipeStep) -> None:
        previous_recipe = self._recipe_versions[-1]
        self._recipe_versions.append(previous_recipe.append_step(recipe_step))

    @property
    def latest_recipe(self) -> AttributeRecipe:  # noqa: D102
        return self._recipe_versions[-1]

    @override
    def append_edge(self, edge: SemanticGraphEdge, weight: int) -> None:
        """Add an edge with the given weight to this path."""
        if self.is_empty:
            self._append_step(edge.tail_node.recipe_step_to_append)
        self._append_step(edge.recipe_step_to_append)
        self._append_step(edge.head_node.recipe_step_to_append)

        super().append_edge(edge, weight)

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        return format_context.formatter.pretty_format(self._nodes)

    @override
    def pop_end(self) -> None:
        if self._edges:
            self._recipe_versions.pop()
            self._recipe_versions.pop()
        else:
            self._recipe_versions.pop()
        super().pop_end()

    @override
    def copy(self) -> AttributeRecipeWriterPath:
        return AttributeRecipeWriterPath(
            _nodes=self._nodes.copy(),
            _edges=self._edges.copy(),
            _current_weight=self._current_weight,
            _current_node_set=self._current_node_set.copy(),
            _weight_addition_order=self._weight_addition_order.copy(),
            _node_set_addition_order=self._node_set_addition_order.copy(),
            _recipe_versions=self._recipe_versions.copy(),
        )


RecipeWriterPathfinder = MetricFlowPathfinder[SemanticGraphNode, SemanticGraphEdge, AttributeRecipeWriterPath]



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from metricflow_semantics.experimental.mf_graph.mf_graph import MetricFlowGraph

logger = logging.getLogger(__name__)


OutputGraphT = TypeVar("OutputGraphT")


class MetricFlowGraphConverter(Generic[OutputGraphT], ABC):
    """Base class for a class that converts graphs."""

    @abstractmethod
    def convert_graph(self, graph: MetricFlowGraph) -> OutputGraphT:
        """Convert the graph to the given output type."""
        raise NotImplementedError()



from __future__ import annotations

import itertools
import logging
import time
from collections import defaultdict
from collections.abc import Set
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.references import MetricReference, SemanticModelReference

from metricflow_semantics.experimental.metricflow_exception import InvalidManifestException
from metricflow_semantics.experimental.ordered_set import MutableOrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.annotated_spec_linkable_element_set import (
    GroupByItemSet,
)
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.mf_logging.runtime import log_runtime
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.model.semantic_model_derivation import SemanticModelDerivation
from metricflow_semantics.model.semantics.linkable_element_set_base import BaseGroupByItemSet
from metricflow_semantics.naming.metric_scheme import MetricNamingScheme
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_pattern_factory import (
    WhereFilterPatternFactory,
)
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolutionLookUp,
)
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_resolver import (
    WhereFilterSpecResolver,
)
from metricflow_semantics.query.group_by_item.group_by_item_resolver import GroupByItemResolution, GroupByItemResolver
from metricflow_semantics.query.group_by_item.resolution_dag.dag import GroupByItemResolutionDag
from metricflow_semantics.query.group_by_item.resolution_dag.dag_builder import GroupByItemResolutionDagBuilder
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.query_resolution_node import (
    QueryGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryResolutionIssueSet,
)
from metricflow_semantics.query.issues.parsing.invalid_apply_group_by import InvalidApplyGroupByIssue
from metricflow_semantics.query.issues.parsing.invalid_limit import InvalidLimitIssue
from metricflow_semantics.query.issues.parsing.invalid_metric import InvalidMetricIssue
from metricflow_semantics.query.issues.parsing.invalid_min_max_only import InvalidMinMaxOnlyIssue
from metricflow_semantics.query.issues.parsing.invalid_order import InvalidOrderByItemIssue
from metricflow_semantics.query.issues.parsing.no_metric_or_group_by import NoMetricOrGroupByIssue
from metricflow_semantics.query.query_resolution import (
    InputToIssueSetMapping,
    InputToIssueSetMappingItem,
    MetricFlowQueryResolution,
)
from metricflow_semantics.query.resolver_inputs.query_resolver_inputs import (
    ResolverInputForApplyGroupBy,
    ResolverInputForGroupByItem,
    ResolverInputForLimit,
    ResolverInputForMetric,
    ResolverInputForMinMaxOnly,
    ResolverInputForOrderByItem,
    ResolverInputForQuery,
    ResolverInputForQueryLevelWhereFilterIntersection,
    ResolverInputForWhereFilterIntersection,
)
from metricflow_semantics.query.suggestion_generator import QueryItemSuggestionGenerator
from metricflow_semantics.query.validation_rules.duplicate_metric import DuplicateMetricValidationRule
from metricflow_semantics.query.validation_rules.metric_time_requirements import MetricTimeQueryValidationRule
from metricflow_semantics.query.validation_rules.query_validator import PostResolutionQueryValidator
from metricflow_semantics.query.validation_rules.unique_column_names import UniqueOutputColumnValidationRule
from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.specs.order_by_spec import OrderBySpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.query_spec import MetricFlowQuerySpec
from metricflow_semantics.specs.spec_set import group_specs_by_type
from metricflow_semantics.workarounds.reference import sorted_semantic_model_references

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ResolveOrderByResult:
    """Result of resolving order-by inputs to specs."""

    order_by_specs: Tuple[OrderBySpec, ...]
    input_to_issue_set_mapping: InputToIssueSetMapping


@dataclass(frozen=True)
class ResolveMetricsResult:
    """Result of resolving a metric input to specs."""

    metric_specs: Tuple[MetricSpec, ...]
    input_to_issue_set_mapping: InputToIssueSetMapping


@dataclass(frozen=True)
class ResolveLimitResult:
    """Result of resolving a limit input."""

    limit: Optional[int]
    input_to_issue_set_mapping: InputToIssueSetMapping


@dataclass(frozen=True)
class ResolveMinMaxOnlyResult:
    """Result of resolving a limit input."""

    min_max_only: bool
    input_to_issue_set_mapping: InputToIssueSetMapping


@dataclass(frozen=True)
class ResolveApplyGroupByResult:
    """Result of resolving a apply group by input."""

    apply_group_by: bool
    input_to_issue_set_mapping: InputToIssueSetMapping


@dataclass(frozen=True)
class ResolveGroupByItemsResult:
    """Result of resolving group-by inputs to specs."""

    resolution_dag: GroupByItemResolutionDag
    group_by_item_specs: Tuple[LinkableInstanceSpec, ...]
    input_to_issue_set_mapping: InputToIssueSetMapping
    linkable_element_set: BaseGroupByItemSet


@dataclass(frozen=True)
class ResolveMetricOrGroupByItemsResult:
    """Result of checking that there are metrics or group by items in the query."""

    input_to_issue_set_mapping: InputToIssueSetMapping


class MetricFlowQueryResolver:
    """Resolves inputs to a query (e.g. metrics, group by items into concrete specs."""

    def __init__(  # noqa: D107
        self,
        manifest_lookup: SemanticManifestLookup,
        where_filter_pattern_factory: WhereFilterPatternFactory,
    ) -> None:
        self._manifest_lookup = manifest_lookup
        self._post_resolution_query_validator = PostResolutionQueryValidator()
        self._where_filter_pattern_factory = where_filter_pattern_factory

    @staticmethod
    def _resolve_has_metric_or_group_by_inputs(
        resolver_input_for_query: ResolverInputForQuery, query_resolution_path: MetricFlowQueryResolutionPath
    ) -> ResolveMetricOrGroupByItemsResult:
        if len(resolver_input_for_query.metric_inputs) == 0 and len(resolver_input_for_query.group_by_item_inputs) == 0:
            return ResolveMetricOrGroupByItemsResult(
                input_to_issue_set_mapping=InputToIssueSetMapping.from_one_item(
                    resolver_input=resolver_input_for_query,
                    issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                        NoMetricOrGroupByIssue.from_parameters(
                            resolver_input_for_query=resolver_input_for_query,
                            query_resolution_path=query_resolution_path,
                        ),
                    ),
                )
            )
        return ResolveMetricOrGroupByItemsResult(input_to_issue_set_mapping=InputToIssueSetMapping.empty_instance())

    @staticmethod
    def _resolve_group_by_item_input(
        group_by_item_input: ResolverInputForGroupByItem,
        group_by_item_resolver: GroupByItemResolver,
    ) -> GroupByItemResolution:
        suggestion_generator = QueryItemSuggestionGenerator(
            input_naming_scheme=group_by_item_input.input_obj_naming_scheme,
            input_str=str(group_by_item_input.input_obj),
            candidate_filters=QueryItemSuggestionGenerator.GROUP_BY_ITEM_CANDIDATE_FILTERS,
        )
        resolution = group_by_item_resolver.resolve_matching_item_for_querying(
            spec_pattern=group_by_item_input.spec_pattern,
            suggestion_generator=suggestion_generator,
        )
        if group_by_item_input.alias:
            resolution = resolution.with_alias(group_by_item_input.alias)
        return resolution

    def _resolve_metric_input(
        self, spec_pattern: SpecPattern, available_metric_specs: Sequence[MetricSpec]
    ) -> List[MetricSpec]:
        start_time = time.perf_counter()
        matching_specs = set(spec_pattern.match(available_metric_specs))

        filtered_metric_specs: List[MetricSpec] = []
        for metric_spec in available_metric_specs:
            if metric_spec in matching_specs:
                filtered_metric_specs.append(metric_spec)

        logger.debug(LazyFormat(lambda: f"Filtering valid metrics took: {time.perf_counter() - start_time:.2f}s"))
        return filtered_metric_specs

    def _resolve_metric_inputs(
        self,
        metric_inputs: Sequence[ResolverInputForMetric],
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> ResolveMetricsResult:
        """Resolve the metric specs associated with the given inputs.

        The order of outputs should be the same as the order of inputs.
        """
        # Build a list of metrics that are available from the manifest.
        available_metric_specs = tuple(
            MetricSpec.from_reference(metric_reference)
            for metric_reference in self._manifest_lookup.metric_lookup.metric_references
        )
        metric_specs: List[MetricSpec] = []
        input_to_issue_set_mapping_items: List[InputToIssueSetMappingItem] = []
        alias_to_metrics: Dict[str, List[Tuple[ResolverInputForMetric, MetricReference]]] = defaultdict(list)

        # Find the metric that matches the metric pattern from the input.
        for metric_input in metric_inputs:
            matching_specs = self._resolve_metric_input(
                spec_pattern=metric_input.spec_pattern, available_metric_specs=available_metric_specs
            )
            if len(matching_specs) == 1:
                matching_spec = matching_specs[0]
                alias = metric_input.alias
                if alias:
                    matching_spec = matching_spec.with_alias(alias)
                metric_specs.append(matching_spec)
                resolved_name = matching_spec.alias or matching_spec.dunder_name
                alias_to_metrics[resolved_name].append((metric_input, matching_spec.reference))
            else:
                suggestion_generator = QueryItemSuggestionGenerator(
                    input_naming_scheme=MetricNamingScheme(),
                    input_str=str(metric_input.input_obj),
                    candidate_filters=(),
                )
                metric_suggestions = suggestion_generator.input_suggestions(candidate_specs=available_metric_specs)
                input_to_issue_set_mapping_items.append(
                    InputToIssueSetMappingItem(
                        resolver_input=metric_input,
                        issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                            InvalidMetricIssue.from_parameters(
                                metric_suggestions=metric_suggestions,
                                query_resolution_path=query_resolution_path,
                            )
                        ),
                    )
                )

        return ResolveMetricsResult(
            metric_specs=tuple(metric_specs),
            input_to_issue_set_mapping=InputToIssueSetMapping(items=tuple(input_to_issue_set_mapping_items)),
        )

    def _resolve_group_by_items_result(
        self,
        metric_references: Sequence[MetricReference],
        group_by_item_inputs: Sequence[ResolverInputForGroupByItem],
        filter_input: ResolverInputForQueryLevelWhereFilterIntersection,
    ) -> ResolveGroupByItemsResult:
        """Resolve the group-by-item specs associated with the given inputs.

        The order of items in the result should be the same as the order of inputs.
        """
        resolution_dag_builder = GroupByItemResolutionDagBuilder(
            manifest_lookup=self._manifest_lookup,
        )
        resolution_dag = resolution_dag_builder.build(
            metric_references=metric_references,
            where_filter_intersection=filter_input.where_filter_intersection,
        )
        logger.debug(LazyFormat(lambda: f"Resolution DAG is:\n{resolution_dag.structure_text()}"))

        group_by_item_resolver = GroupByItemResolver(
            manifest_lookup=self._manifest_lookup,
            resolution_dag=resolution_dag,
        )

        input_to_issue_set_mapping_items: List[InputToIssueSetMappingItem] = []
        group_by_item_specs: List[LinkableInstanceSpec] = []
        linkable_element_sets: List[BaseGroupByItemSet] = []

        for group_by_item_input in group_by_item_inputs:
            resolution = MetricFlowQueryResolver._resolve_group_by_item_input(
                group_by_item_resolver=group_by_item_resolver,
                group_by_item_input=group_by_item_input,
            )
            if resolution.issue_set.has_issues:
                input_to_issue_set_mapping_items.append(
                    InputToIssueSetMappingItem(resolver_input=group_by_item_input, issue_set=resolution.issue_set)
                )
            if resolution.spec is not None:
                group_by_item_specs.append(resolution.spec)
                linkable_element_sets.append(resolution.linkable_element_set)

        linkable_element_set: BaseGroupByItemSet
        if len(linkable_element_sets) == 0:
            linkable_element_set = GroupByItemSet()
        else:
            linkable_element_set = linkable_element_sets[0].union(*linkable_element_sets[1:])

        return ResolveGroupByItemsResult(
            resolution_dag=resolution_dag,
            group_by_item_specs=tuple(group_by_item_specs),
            input_to_issue_set_mapping=InputToIssueSetMapping(tuple(input_to_issue_set_mapping_items)),
            linkable_element_set=linkable_element_set,
        )

    @staticmethod
    def _resolve_order_by(
        resolver_inputs_for_order_by_items: Sequence[ResolverInputForOrderByItem],
        metric_specs: Sequence[MetricSpec],
        group_by_item_specs: Sequence[LinkableInstanceSpec],
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> ResolveOrderByResult:
        mapping_items: List[InputToIssueSetMappingItem] = []
        order_by_specs: List[OrderBySpec] = []

        # Match the pattern from the order by input to one of the metric or group-by-item specs.
        # The pattern needs to be used because there are cases where the order-by-item is specified in a different way
        # from the group-by-item, so an equality comparison won't work.
        for resolver_input_for_order_by in resolver_inputs_for_order_by_items:
            matching_specs: set[InstanceSpec] = set()
            for possible_input in resolver_input_for_order_by.possible_inputs:
                spec_pattern = possible_input.spec_pattern
                matching_specs.update(spec_pattern.match(metric_specs))
                matching_specs.update(spec_pattern.match(group_by_item_specs))

            if len(matching_specs) != 1:
                mapping_items.append(
                    InputToIssueSetMappingItem(
                        resolver_input=resolver_input_for_order_by,
                        issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                            InvalidOrderByItemIssue.from_parameters(
                                order_by_item_input=resolver_input_for_order_by,
                                query_resolution_path=query_resolution_path,
                            )
                        ),
                    )
                )
            else:
                order_by_specs.append(
                    OrderBySpec(
                        # Ignore aliases in the order by since we'll render the expression instead of the alias.
                        instance_spec=matching_specs.pop().with_alias(None),
                        descending=resolver_input_for_order_by.descending,
                    )
                )

        return ResolveOrderByResult(
            input_to_issue_set_mapping=InputToIssueSetMapping(items=tuple(mapping_items)),
            order_by_specs=tuple(order_by_specs),
        )

    @staticmethod
    def _resolve_limit_input(
        limit_input: ResolverInputForLimit, query_resolution_path: MetricFlowQueryResolutionPath
    ) -> ResolveLimitResult:
        limit = limit_input.limit
        if limit is not None and limit < 0:
            return ResolveLimitResult(
                limit=limit,
                input_to_issue_set_mapping=InputToIssueSetMapping.from_one_item(
                    resolver_input=limit_input,
                    issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                        InvalidLimitIssue.from_parameters(limit=limit, query_resolution_path=query_resolution_path),
                    ),
                ),
            )
        return ResolveLimitResult(limit=limit, input_to_issue_set_mapping=InputToIssueSetMapping.empty_instance())

    @staticmethod
    def _resolve_min_max_only_input(
        min_max_only_input: ResolverInputForMinMaxOnly,
        query_resolution_path: MetricFlowQueryResolutionPath,
        metric_inputs: Tuple[ResolverInputForMetric, ...],
        group_by_item_inputs: Tuple[ResolverInputForGroupByItem, ...],
        order_by_item_inputs: Tuple[ResolverInputForOrderByItem, ...],
        limit_input: ResolverInputForLimit,
    ) -> ResolveMinMaxOnlyResult:
        min_max_only = min_max_only_input.min_max_only
        if min_max_only:
            if (
                metric_inputs
                or order_by_item_inputs
                or (limit_input.limit is not None)
                or len(group_by_item_inputs) != 1
            ):
                return ResolveMinMaxOnlyResult(
                    min_max_only=min_max_only,
                    input_to_issue_set_mapping=InputToIssueSetMapping.from_one_item(
                        resolver_input=min_max_only_input,
                        issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                            InvalidMinMaxOnlyIssue.from_parameters(
                                min_max_only=min_max_only, query_resolution_path=query_resolution_path
                            ),
                        ),
                    ),
                )
        return ResolveMinMaxOnlyResult(
            min_max_only=min_max_only, input_to_issue_set_mapping=InputToIssueSetMapping.empty_instance()
        )

    @staticmethod
    def _resolve_apply_group_by_input(
        apply_group_by_input: ResolverInputForApplyGroupBy,
        metric_inputs: Tuple[ResolverInputForMetric, ...],
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> ResolveApplyGroupByResult:
        if metric_inputs and not apply_group_by_input.apply_group_by:
            return ResolveApplyGroupByResult(
                apply_group_by=apply_group_by_input.apply_group_by,
                input_to_issue_set_mapping=InputToIssueSetMapping.from_one_item(
                    resolver_input=apply_group_by_input,
                    issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                        InvalidApplyGroupByIssue.from_parameters(
                            apply_group_by=apply_group_by_input.apply_group_by,
                            metric_inputs=metric_inputs,
                            query_resolution_path=query_resolution_path,
                        ),
                    ),
                ),
            )
        return ResolveApplyGroupByResult(
            apply_group_by=apply_group_by_input.apply_group_by,
            input_to_issue_set_mapping=InputToIssueSetMapping.empty_instance(),
        )

    def _build_filter_spec_lookup(
        self,
        resolution_dag: GroupByItemResolutionDag,
    ) -> FilterSpecResolutionLookUp:
        where_filter_spec_resolver = WhereFilterSpecResolver(
            manifest_lookup=self._manifest_lookup,
            resolution_dag=resolution_dag,
            spec_pattern_factory=self._where_filter_pattern_factory,
        )

        return where_filter_spec_resolver.resolve_lookup()

    def resolve_query(self, resolver_input_for_query: ResolverInputForQuery) -> MetricFlowQueryResolution:
        """Resolve the query into specs that can be passed into the next stage in query processing."""
        # Workaround for a Pycharm type inspection issue with decorators.
        # noinspection PyArgumentList
        return self._resolve_query(resolver_input_for_query=resolver_input_for_query)

    @log_runtime()
    def _resolve_query(self, resolver_input_for_query: ResolverInputForQuery) -> MetricFlowQueryResolution:
        metric_inputs = resolver_input_for_query.metric_inputs
        group_by_item_inputs = resolver_input_for_query.group_by_item_inputs
        order_by_item_inputs = resolver_input_for_query.order_by_item_inputs
        limit_input = resolver_input_for_query.limit_input
        query_level_filter_input = resolver_input_for_query.filter_input
        min_max_only_input = resolver_input_for_query.min_max_only
        apply_group_by_input = resolver_input_for_query.apply_group_by

        mappings_to_merge: List[InputToIssueSetMapping] = []

        # Resolve metrics.
        resolve_metrics_result = self._resolve_metric_inputs(
            metric_inputs=metric_inputs,
            query_resolution_path=MetricFlowQueryResolutionPath.empty_instance(),
        )
        mappings_to_merge.append(resolve_metrics_result.input_to_issue_set_mapping)
        metric_specs = resolve_metrics_result.metric_specs

        # Define a resolution path for issues where the input is considered to be the whole query.
        query_resolution_path = MetricFlowQueryResolutionPath.from_path_item(
            QueryGroupByItemResolutionNode.create(
                parent_nodes=(),
                metrics_in_query=tuple(metric_spec.reference for metric_spec in metric_specs),
                where_filter_intersection=query_level_filter_input.where_filter_intersection,
            )
        )

        # Check that the query contains metrics or group by items.
        resolve_metrics_or_group_by_result = self._resolve_has_metric_or_group_by_inputs(
            resolver_input_for_query=resolver_input_for_query, query_resolution_path=query_resolution_path
        )
        mappings_to_merge.append(resolve_metrics_or_group_by_result.input_to_issue_set_mapping)

        # Resolve limit
        resolve_limit_result = self._resolve_limit_input(
            limit_input=limit_input,
            query_resolution_path=query_resolution_path,
        )
        mappings_to_merge.append(resolve_limit_result.input_to_issue_set_mapping)

        # Resolve min max only
        resolve_min_max_only_result = self._resolve_min_max_only_input(
            min_max_only_input=min_max_only_input,
            query_resolution_path=query_resolution_path,
            metric_inputs=metric_inputs,
            group_by_item_inputs=group_by_item_inputs,
            order_by_item_inputs=order_by_item_inputs,
            limit_input=limit_input,
        )
        mappings_to_merge.append(resolve_min_max_only_result.input_to_issue_set_mapping)

        # Resolve apply group by
        resolve_apply_group_by_result = self._resolve_apply_group_by_input(
            apply_group_by_input=apply_group_by_input,
            metric_inputs=metric_inputs,
            query_resolution_path=query_resolution_path,
        )
        mappings_to_merge.append(resolve_apply_group_by_result.input_to_issue_set_mapping)

        # Early stop before resolving further as with invalid metrics, the errors won't be as useful.
        issue_set_mapping_so_far = InputToIssueSetMapping.merge_iterable(mappings_to_merge)
        if issue_set_mapping_so_far.has_issues:
            return MetricFlowQueryResolution(
                query_spec=None,
                resolution_dag=None,
                filter_spec_lookup=FilterSpecResolutionLookUp.empty_instance(),
                input_to_issue_set=issue_set_mapping_so_far,
                queried_semantic_models=(),
            )

        # Resolve group by items.
        resolve_group_by_item_result = self._resolve_group_by_items_result(
            metric_references=tuple(metric_spec.reference for metric_spec in metric_specs),
            group_by_item_inputs=group_by_item_inputs,
            filter_input=query_level_filter_input,
        )
        resolution_dag = resolve_group_by_item_result.resolution_dag
        group_by_item_specs = resolve_group_by_item_result.group_by_item_specs
        mappings_to_merge.append(resolve_group_by_item_result.input_to_issue_set_mapping)

        # Resolve order by.
        resolve_order_by_result = MetricFlowQueryResolver._resolve_order_by(
            resolver_inputs_for_order_by_items=order_by_item_inputs,
            metric_specs=metric_specs,
            group_by_item_specs=group_by_item_specs,
            query_resolution_path=query_resolution_path,
        )
        order_by_specs = resolve_order_by_result.order_by_specs
        if resolve_order_by_result.input_to_issue_set_mapping.has_issues:
            mappings_to_merge.append(resolve_order_by_result.input_to_issue_set_mapping)

        # Resolve all where filters in the DAG and generate mappings if there are issues.
        filter_spec_lookup = self._build_filter_spec_lookup(resolution_dag)
        for filter_spec_resolution in filter_spec_lookup.spec_resolutions:
            if filter_spec_resolution.issue_set.has_issues:
                mappings_to_merge.append(
                    InputToIssueSetMapping.from_one_item(
                        resolver_input=ResolverInputForWhereFilterIntersection(
                            filter_resolution_path=filter_spec_resolution.filter_location_path,
                            where_filter_intersection=filter_spec_resolution.where_filter_intersection,
                            object_builder_str=filter_spec_resolution.object_builder_str,
                        ),
                        issue_set=filter_spec_resolution.issue_set,
                    )
                )

        for non_parseable_resolution in filter_spec_lookup.non_parsable_resolutions:
            filter_resolver_input = ResolverInputForWhereFilterIntersection(
                filter_resolution_path=non_parseable_resolution.filter_location_path,
                where_filter_intersection=non_parseable_resolution.where_filter_intersection,
                object_builder_str=None,
            )
            mappings_to_merge.append(
                InputToIssueSetMapping.from_one_item(
                    resolver_input=filter_resolver_input,
                    issue_set=non_parseable_resolution.issue_set,
                )
            )

        # Return if there are any errors as issues generated by later validations may not make sense otherwise.
        issue_set_mapping_so_far = InputToIssueSetMapping.merge_iterable(mappings_to_merge)
        if issue_set_mapping_so_far.has_issues:
            return MetricFlowQueryResolution(
                query_spec=None,
                resolution_dag=resolution_dag,
                filter_spec_lookup=filter_spec_lookup,
                input_to_issue_set=issue_set_mapping_so_far,
                queried_semantic_models=(),
            )

        # No errors.
        linkable_spec_set = group_specs_by_type(group_by_item_specs)
        logger.debug(
            LazyFormat(lambda: f"Group-by-items were resolved to:\n{mf_pformat(linkable_spec_set.linkable_specs)}")
        )

        # Run post-resolution validation rules to generate issues that are generated at the query-level.
        query_level_issue_set = self._post_resolution_query_validator.validate_query(
            resolution_dag=resolution_dag,
            resolver_input_for_query=resolver_input_for_query,
            validation_rules=(
                MetricTimeQueryValidationRule(
                    manifest_lookup=self._manifest_lookup,
                    resolver_input_for_query=resolver_input_for_query,
                    resolve_group_by_item_result=resolve_group_by_item_result,
                    resolve_metric_result=resolve_metrics_result,
                ),
                DuplicateMetricValidationRule(
                    manifest_lookup=self._manifest_lookup,
                    resolver_input_for_query=resolver_input_for_query,
                    resolve_group_by_item_result=resolve_group_by_item_result,
                    resolve_metric_result=resolve_metrics_result,
                ),
                UniqueOutputColumnValidationRule(
                    manifest_lookup=self._manifest_lookup,
                    resolver_input_for_query=resolver_input_for_query,
                    resolve_group_by_item_result=resolve_group_by_item_result,
                    resolve_metric_result=resolve_metrics_result,
                ),
            ),
        )

        if query_level_issue_set.has_issues:
            mappings_to_merge.append(
                InputToIssueSetMapping(
                    (
                        InputToIssueSetMappingItem(
                            resolver_input=resolver_input_for_query,
                            issue_set=query_level_issue_set,
                        ),
                    )
                )
            )

        issue_set_mapping = InputToIssueSetMapping.merge_iterable(mappings_to_merge)

        if issue_set_mapping.has_issues:
            return MetricFlowQueryResolution(
                query_spec=None,
                resolution_dag=resolution_dag,
                filter_spec_lookup=filter_spec_lookup,
                input_to_issue_set=issue_set_mapping,
                queried_semantic_models=(),
            )

        model_reference_set = set(resolve_group_by_item_result.linkable_element_set.derived_from_semantic_models)
        for filter_spec_resolution in filter_spec_lookup.spec_resolutions:
            model_reference_set.update(
                set(filter_spec_resolution.resolved_group_by_item_set.derived_from_semantic_models)
            )

        # Collect all semantic models referenced by the query.
        semantic_models_in_group_by_items = set(
            resolve_group_by_item_result.linkable_element_set.derived_from_semantic_models
        )
        semantic_models_in_filters = set(
            itertools.chain.from_iterable(
                filter_spec_resolution.resolved_group_by_item_set.derived_from_semantic_models
                for filter_spec_resolution in filter_spec_lookup.spec_resolutions
            )
        )
        simple_metric_semantic_models = self._get_models_for_simple_metrics(resolution_dag)

        queried_semantic_models = set.union(
            semantic_models_in_group_by_items, semantic_models_in_filters, simple_metric_semantic_models
        )
        queried_semantic_models -= {SemanticModelDerivation.VIRTUAL_SEMANTIC_MODEL_REFERENCE}

        # Sanity check to make sure that all queried semantic models are in the model.
        models_not_in_manifest = queried_semantic_models - {
            semantic_model.reference for semantic_model in self._manifest_lookup.semantic_manifest.semantic_models
        }

        # There are no known cases where this should happen, but adding this check just in case there's a bug where
        # a metric alias is used incorrectly.
        if len(models_not_in_manifest) > 0:
            logger.error(
                LazyFormat(
                    "Semantic references that aren't in the manifest were found in the set used in"
                    " a query. This is a bug, and to avoid potential issues, they will be filtered out.",
                    models_not_in_manifest=models_not_in_manifest,
                )
            )
        queried_semantic_models -= models_not_in_manifest

        return MetricFlowQueryResolution(
            query_spec=MetricFlowQuerySpec(
                metric_specs=metric_specs,
                dimension_specs=linkable_spec_set.dimension_specs,
                entity_specs=linkable_spec_set.entity_specs,
                time_dimension_specs=linkable_spec_set.time_dimension_specs,
                order_by_specs=tuple(order_by_specs),
                limit=limit_input.limit,
                filter_intersection=query_level_filter_input.where_filter_intersection,
                filter_spec_resolution_lookup=filter_spec_lookup,
                min_max_only=min_max_only_input.min_max_only,
                apply_group_by=apply_group_by_input.apply_group_by,
                spec_output_order=group_by_item_specs + metric_specs,
            ),
            resolution_dag=resolution_dag,
            filter_spec_lookup=filter_spec_lookup,
            input_to_issue_set=issue_set_mapping,
            queried_semantic_models=sorted_semantic_model_references(queried_semantic_models),
        )

    def _get_models_for_simple_metrics(self, resolution_dag: GroupByItemResolutionDag) -> Set[SemanticModelReference]:
        """Return the semantic model references for all simple metrics used in the query."""
        resolution_dag_node_set = resolution_dag.sink_node.inclusive_ancestors()

        simple_metric_references: MutableOrderedSet[MetricReference] = MutableOrderedSet()

        # Collect simple metrics.
        for simple_metric_node in resolution_dag_node_set.simple_metric_nodes:
            simple_metric_references.add(simple_metric_node.metric_reference)

        # For conversion metrics, get the input conversion metric through the metric since those
        # aren't in the DAG.
        for metric_node in resolution_dag_node_set.complex_metric_nodes:
            metric = self._manifest_lookup.metric_lookup.get_metric(metric_node.metric_reference)
            conversion_type_params = metric.type_params.conversion_type_params
            if conversion_type_params is None:
                continue

            assert conversion_type_params.base_metric is not None, "A conversion metric must have a base metric."
            assert (
                conversion_type_params.conversion_metric is not None
            ), "A conversion metric must have an input conversion metric."
            # The input base metric should be in the DAG, but just in case.
            simple_metric_references.add(MetricReference(conversion_type_params.base_metric.name))
            simple_metric_references.add(MetricReference(conversion_type_params.conversion_metric.name))

        model_references: MutableOrderedSet[SemanticModelReference] = MutableOrderedSet()
        for simple_metric_reference in simple_metric_references:
            metric = self._manifest_lookup.metric_lookup.get_metric(simple_metric_reference)
            metric_aggregation_params = metric.type_params.metric_aggregation_params
            if metric_aggregation_params is None:
                raise InvalidManifestException(
                    LazyFormat("Metric does not have `metric_aggregation_params` set", metric=metric)
                )
            model_references.add(SemanticModelReference(metric_aggregation_params.semantic_model))

        return model_references



from __future__ import annotations

import importlib.util
import logging
import threading
from types import ModuleType
from typing import Mapping, Optional, Sequence, Union

import tabulate

logger = logging.getLogger(__name__)


class IsolatedTabulateRunner:
    """Helps run `tabulate` with different module options.

    The `tabulate.tabulate` method uses some options defined in the module instead of being provided as arguments to
    the function. This runner is used to change those options in isolation by loading a copy of the `tabulate` module.
    This helps to ensure that other calls to `tabulate.tabulate` don't see unexpected results.

    The next version of `tabluate` will provide those options as call arguments, so this can be removed then.
    """

    _TABULATE_MODULE_COPY: Optional[ModuleType] = None
    _STATE_LOCK = threading.Lock()

    @classmethod
    def tabulate(
        cls,
        tabular_data: Sequence[Union[Mapping[str, object], Sequence[object]]],
        headers: Union[str, Sequence[str]],
        column_alignment: Optional[Sequence[str]] = None,
        tablefmt: str = "simple",
    ) -> str:
        """Produce a text table from the given data. Also see class docstring."""
        with IsolatedTabulateRunner._STATE_LOCK:
            try:
                if IsolatedTabulateRunner._TABULATE_MODULE_COPY is None:
                    tabulate_module_spec = importlib.util.find_spec("tabulate")
                    if tabulate_module_spec is None:
                        raise RuntimeError("Unable to find spec for `tabulate`.")
                    tabulate_module_copy = importlib.util.module_from_spec(tabulate_module_spec)
                    if tabulate_module_copy is None:
                        raise RuntimeError(f"Unable to load module using {tabulate_module_spec=}")
                    if tabulate_module_spec.loader is None:
                        raise RuntimeError(f"Loader missing for {tabulate_module_spec=}")
                    tabulate_module_spec.loader.exec_module(tabulate_module_copy)
                    tabulate_module_copy.PRESERVE_WHITESPACE = True  # type: ignore[attr-defined]
                    IsolatedTabulateRunner._TABULATE_MODULE_COPY = tabulate_module_copy
            except Exception:
                logger.exception(
                    "Failed to load a copy of the `tabulate` module. This means that some table-formatting options "
                    "can't be applied. This is a bug and should be investigated."
                )

        # `tabulate.tabulate` can detect numeric values and apply special formatting rules. This can
        # result in unexpected values when coupled with the `--decimals` option, so disabling that feature.
        disable_numparse = True

        if IsolatedTabulateRunner._TABULATE_MODULE_COPY is None:
            logger.warning(
                "Generating text table without required options set as there was an error loading the "
                "`tabulate` module."
            )
            return tabulate.tabulate(
                tabular_data=tabular_data,
                headers=headers,
                disable_numparse=disable_numparse,
                colalign=column_alignment,
                tablefmt=tablefmt,
            )

        return IsolatedTabulateRunner._TABULATE_MODULE_COPY.tabulate(
            tabular_data=tabular_data,
            headers=headers,
            disable_numparse=disable_numparse,
            colalign=column_alignment,
            tablefmt=tablefmt,
        )



"""Classes for representing DOT objects as `graphviz` uses strings as arguments.

Generally, only the attributes / attribute values that are used in this project are listed. For a full list, see:
    https://www.graphviz.org/doc/info/attrs.html
"""
from __future__ import annotations

import logging
from collections.abc import Mapping
from enum import Enum
from typing import Optional

from metricflow_semantics.collection_helpers.syntactic_sugar import (
    mf_ensure_mapping,
    mf_first_non_none,
    mf_first_non_none_or_raise,
)
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass

logger = logging.getLogger(__name__)


class DotNodeShape(Enum):
    """An enumeration of the `graphviz` shapes used in this project."""

    PLAIN = "plain"
    BOX = "box"
    ELLIPSE = "ellipse"


class DotEdgeArrowShape(Enum):
    """An enumeration of the `graphviz` edge-arrow shapes used in this project."""

    NORMAL = "normal"
    NONE = "none"


class DotColor(Enum):
    """An enumeration of the `graphviz` colors used in this project."""

    CORNFLOWER_BLUE = "#87CEFA"
    SALMON_PINK = "#FFA07A"
    EXTRA_DIM_GRAY = "#303030"
    GRAY = "#808080"
    DARK_GRAY = "#181818"
    LIGHT_GRAY = "#D3D3D3"
    LIME_GREEN = "#006400"
    BLACK = "#000000"
    GOLD = "#FFD700"
    # Note the 2 additional hex characters for the alpha.
    TRANSPARENT = "#00000000"


@fast_frozen_dataclass()
class DotGraphAttributeSet:
    """Attributes for graphs in DOT.

    Note: this does not contain the objects in the graph.
    """

    # Name of the graph.
    name: str
    # Additional keyword arguments for the call to create a DOT graph instance.
    additional_kwargs: Mapping[str, str]

    @staticmethod
    def create(name: str, additional_kwargs: Optional[Mapping[str, str]] = None) -> DotGraphAttributeSet:
        """See field comments."""
        return DotGraphAttributeSet(
            name=name,
            additional_kwargs=mf_ensure_mapping(additional_kwargs),
        )

    def with_attributes(
        self, name: Optional[str] = None, dot_kwargs: Optional[Mapping[str, str]] = None
    ) -> DotGraphAttributeSet:
        """Create a copy but with fields set to the present arguments."""
        return DotGraphAttributeSet(
            name=mf_first_non_none_or_raise(name, self.name),
            additional_kwargs=dict(self.additional_kwargs, **(mf_ensure_mapping(dot_kwargs))),
        )

    def merge(self, other: DotGraphAttributeSet) -> DotGraphAttributeSet:  # noqa: D102
        return DotGraphAttributeSet(
            name=mf_first_non_none_or_raise(self.name, other.name),
            additional_kwargs=dict(self.additional_kwargs, **(mf_ensure_mapping(other.additional_kwargs))),
        )

    @property
    def dot_graph_attrs(self) -> dict[str, str]:
        """Return the `graph_attr` dict for creating the graph in DOT."""
        kwargs = {"name": self.name}
        kwargs.update(self.additional_kwargs)
        return kwargs


@fast_frozen_dataclass()
class DotEdgeAttributeSet:
    """Attributes for edges in DOT."""

    tail_name: str
    head_name: str
    color: Optional[DotColor]
    label: Optional[str]
    arrow_shape: Optional[DotEdgeArrowShape]
    additional_kwargs: Mapping[str, str]

    @staticmethod
    def create(  # noqa: D102
        tail_name: str,
        head_name: str,
        color: Optional[DotColor] = None,
        label: Optional[str] = None,
        arrow_shape: Optional[DotEdgeArrowShape] = None,
        additional_kwargs: Optional[Mapping[str, str]] = None,
    ) -> DotEdgeAttributeSet:
        return DotEdgeAttributeSet(
            tail_name=tail_name,
            head_name=head_name,
            color=color,
            label=label,
            arrow_shape=arrow_shape,
            additional_kwargs=mf_ensure_mapping(additional_kwargs),
        )

    def merge(self, other: DotEdgeAttributeSet) -> DotEdgeAttributeSet:  # noqa: D102
        return DotEdgeAttributeSet(
            tail_name=other.tail_name,
            head_name=other.head_name,
            color=mf_first_non_none(other.color, self.color),
            label=mf_first_non_none(other.label, self.label),
            arrow_shape=mf_first_non_none(other.arrow_shape, self.arrow_shape),
            additional_kwargs=dict(**self.additional_kwargs, **other.additional_kwargs),
        )

    @property
    def dot_kwargs(self) -> Mapping[str, str]:
        """Return the keyword arguments for calling `.edge()` in DOT."""
        kwargs = {
            "tail_name": self.tail_name,
            "head_name": self.head_name,
        }
        if self.color is not None:
            kwargs["color"] = self.color.value

        if self.label is not None:
            kwargs["label"] = self.label

        if self.arrow_shape is not None:
            kwargs["arrowhead"] = self.arrow_shape.value

        kwargs.update(**self.additional_kwargs)

        return kwargs


class DotRankKey(Enum):
    """A key used to group nodes and assign them the same `graphviz` rank.

    Only edge-as-nodes are configured currently, but more nodes may be tagged.
    """

    EDGE_AS_NODE = "edge"


@fast_frozen_dataclass()
class DotNodeAttributeSet:
    """Attributes for nodes in DOT."""

    name: str
    color: Optional[DotColor]
    fill_color: Optional[DotColor]
    label: Optional[str]
    shape: Optional[DotNodeShape]

    # When rendering graphs, nodes with the same rank key are force to have the same rank in the rendering engine.
    rank_key: Optional[DotRankKey]

    # When rendering an edge `A -> B` as `A -> [Label Node] -> B`, the label node can be placed in a more organized
    # way if it is place in the same cluster as the one for `A` or `B`. Since the nodes could belong to different
    # clusters the cluster of the node with the highest priority is selected.
    cluster_priority_for_edge_as_node: Optional[int]

    # Additional keyword arguments for the call to create a DOT graph instance.
    additional_kwargs: Mapping[str, str]

    @staticmethod
    def create(  # noqa: D102
        name: str,
        color: Optional[DotColor] = None,
        fill_color: Optional[DotColor] = None,
        label: Optional[str] = None,
        shape: Optional[DotNodeShape] = None,
        rank_key: Optional[DotRankKey] = None,
        edge_node_priority: Optional[int] = None,
        additional_kwargs: Optional[Mapping[str, str]] = None,
    ) -> DotNodeAttributeSet:
        return DotNodeAttributeSet(
            name=name,
            color=color,
            fill_color=fill_color,
            label=label,
            shape=shape,
            rank_key=rank_key,
            cluster_priority_for_edge_as_node=edge_node_priority,
            additional_kwargs=mf_ensure_mapping(additional_kwargs),
        )

    def merge(self, other: DotNodeAttributeSet) -> DotNodeAttributeSet:  # noqa: D102
        return DotNodeAttributeSet.create(
            name=other.name,
            color=other.color or self.color,
            label=other.label or self.label,
            shape=other.shape or self.shape,
            rank_key=other.rank_key or self.rank_key,
            edge_node_priority=other.cluster_priority_for_edge_as_node or self.cluster_priority_for_edge_as_node,
            additional_kwargs=dict(**self.additional_kwargs, **other.additional_kwargs),
        )

    @property
    def dot_kwargs(self) -> Mapping[str, str]:
        """Return the keyword arguments for calling `.node()` in DOT."""
        kwargs = {
            "name": self.name,
        }
        if self.color is not None:
            kwargs["color"] = self.color.value

        if self.fill_color is not None:
            kwargs["fillcolor"] = self.fill_color.value

        if self.label is not None:
            kwargs["label"] = self.label

        if self.shape is not None:
            kwargs["shape"] = self.shape.value

        kwargs.update(**self.additional_kwargs)
        return kwargs

    def with_attributes(
        self,
        color: Optional[DotColor] = None,
        fill_color: Optional[DotColor] = None,
        label: Optional[str] = None,
        shape: Optional[DotNodeShape] = None,
        rank_key: Optional[DotRankKey] = None,
        edge_node_priority: Optional[int] = None,
        additional_kwargs: Optional[Mapping[str, str]] = None,
    ) -> DotNodeAttributeSet:
        """Return a copy of this with fields set to the present arguments."""
        return DotNodeAttributeSet(
            name=self.name,
            color=mf_first_non_none(color, self.color),
            fill_color=mf_first_non_none(fill_color, self.fill_color),
            label=mf_first_non_none(label, self.label),
            shape=mf_first_non_none(shape, self.shape),
            rank_key=mf_first_non_none(rank_key, self.rank_key),
            cluster_priority_for_edge_as_node=edge_node_priority,
            additional_kwargs=mf_ensure_mapping(mf_first_non_none(additional_kwargs, self.additional_kwargs)),
        )



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Sequence

from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraphEdge,
)

logger = logging.getLogger(__name__)


class SemanticSubgraphGenerator(ABC):
    """Generates a specific portion of the semantic graph.

    As the semantic graph has many elements, it's helpful to break down the generation of the graph into logical pieces.
    For example, the `TimeEntitySubgraphGenerator` creates a subgraph containing the nodes and edges related to the
    time entities.

    To create the complete semantic graph, the output of each generator is combined into a single graph.

    Only the edges need to be returned as the graph implementation adds the end nodes of the edge as needed.
    """

    def __init__(self, manifest_object_lookup: ManifestObjectLookup) -> None:  # noqa: D107
        self._manifest_object_lookup = manifest_object_lookup

    @abstractmethod
    def add_edges_for_manifest(self, edge_list: list[SemanticGraphEdge]) -> None:
        """Add edges appropriate for the manifest / generator to the `edge_list`.

        This method mutates `edge_list` to reduce the overhead of concatenating edges from multiple generators. This is
        a factor for manifests with high complexity.
        """
        raise NotImplementedError

    def generate_edges(self) -> Sequence[SemanticGraphEdge]:
        """Return a list of edges that should be added to the graph."""
        edge_list: list[SemanticGraphEdge] = []
        self.add_edges_for_manifest(edge_list)
        return edge_list



from __future__ import annotations

import itertools
import logging
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Optional

from typing_extensions import override

from metricflow_semantics.experimental.comparison_helpers import ComparisonOtherType
from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable
from metricflow_semantics.mf_logging.pretty_formatter import PrettyFormatContext

logger = logging.getLogger(__name__)


class MetricFlowGraphId(ABC):
    """Interface for an object to identify different graphs."""

    @property
    @abstractmethod
    def str_value(self) -> str:
        """Return a string representation of the ID."""
        raise NotImplementedError


class SequentialGraphId(MetricFlowGraphId, MetricFlowPrettyFormattable):
    """Graph IDs that are generated sequentially."""

    # `itertools.count()` returns an iterable that is thread-safe, so this is a way of generating sequential
    # IDs without using a lock.
    _ID_COUNTER = itertools.count()

    def __init__(self) -> None:  # noqa: D107
        self._str_value = "id_" + str(next(SequentialGraphId._ID_COUNTER))

    @staticmethod
    def create() -> SequentialGraphId:  # noqa: D102
        """TODO: Remove and update call sites."""
        return SequentialGraphId()

    @override
    @property
    def str_value(self) -> str:
        return self.str_value

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        return self._str_value

    @override
    def __eq__(self, other: ComparisonOtherType) -> bool:
        if self.__class__ is not other.__class__:
            return False

        return self._str_value == other._str_value

    @cached_property
    def _cached_hash(self) -> int:
        return hash(self._str_value)

    @override
    def __hash__(self) -> int:
        return self._cached_hash



from __future__ import annotations

from enum import Enum

from dbt_semantic_interfaces.type_enums.aggregation_type import AggregationType


def fill_nulls_with_0(agg_type: AggregationType) -> bool:
    """Indicates if charts should show 0 instead of null where there are gaps in data."""
    return agg_type in (
        AggregationType.SUM,
        AggregationType.COUNT_DISTINCT,
        AggregationType.SUM_BOOLEAN,
        AggregationType.COUNT,
    )


def can_limit_dimension_values(agg_type: AggregationType) -> bool:
    """Indicates if we can limit dimension values in charts.

    Currently, this means:
    1. The dimensions we care about most are the ones with the highest numeric values
    2. We can calculate the "other" column in the postprocessor (meaning the metric is expansive)
    """
    return agg_type in (AggregationType.SUM, AggregationType.SUM_BOOLEAN, AggregationType.COUNT)


class AggregationState(Enum):
    """Represents how the measure is aggregated."""

    # When reading from the source, the measure is considered non-aggregated.
    NON_AGGREGATED = "NON_AGGREGATED"
    PARTIAL = "PARTIAL"
    # Aggregated to the grain of the group-by-items
    COMPLETE = "COMPLETE"

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}.{self.name}"



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import override

from metricflow.dataflow.builder.aggregation_helper import InstanceAliasMapping, NullFillValueMapping
from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class AggregateSimpleMetricInputsNode(DataflowPlanNode):
    """A node that aggregates the simple-metric inputs by the associated group by elements.

    In the event that one or more of the aggregated input simple-metric inputs has an alias assigned to it, any output query
    resulting from an operation on this node must apply the alias and transform the simple-metric-input instances accordingly,
    otherwise this join could produce a query with two identically named simple-metric-input columns with, e.g., different
    constraints applied to the simple-metric input.

    The simple-metric-input specs are required for downstream nodes to be aware of any simple-metric inputs with
    user-provided aliases, such as we might encounter with constrained and unconstrained versions of the
    same simple-metric.
    """

    alias_mapping: InstanceAliasMapping
    null_fill_value_mapping: NullFillValueMapping

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode,
        alias_mapping: InstanceAliasMapping,
        null_fill_value_mapping: NullFillValueMapping,
    ) -> AggregateSimpleMetricInputsNode:
        return AggregateSimpleMetricInputsNode(
            parent_nodes=(parent_node,),
            alias_mapping=alias_mapping,
            null_fill_value_mapping=null_fill_value_mapping,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_AGGREGATE_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_aggregate_simple_metric_inputs_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return """Aggregate Inputs for Simple Metrics"""

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and self.alias_mapping == other_node.alias_mapping
            and self.null_fill_value_mapping == other_node.null_fill_value_mapping
        )

    def with_new_parents(  # noqa: D102
        self, new_parent_nodes: Sequence[DataflowPlanNode]
    ) -> AggregateSimpleMetricInputsNode:
        assert len(new_parent_nodes) == 1
        return AggregateSimpleMetricInputsNode(
            parent_nodes=tuple(new_parent_nodes),
            alias_mapping=self.alias_mapping,
            null_fill_value_mapping=self.null_fill_value_mapping,
        )

    @cached_property
    @override
    def displayed_properties(self) -> Sequence[DisplayedProperty]:
        properties: list[DisplayedProperty] = list(super().displayed_properties)
        properties.append(DisplayedProperty("alias_mapping", self.alias_mapping.element_name_to_alias))
        properties.append(
            DisplayedProperty("null_fill_value_mapping", self.null_fill_value_mapping.element_name_to_null_fill_value)
        )
        return properties



from __future__ import annotations

import itertools
import logging
import threading
from abc import ABC
from collections import defaultdict
from collections.abc import Set
from typing import Final, Generic, Iterator, Optional, Sequence, TypeVar

from metricflow_semantics.collection_helpers.syntactic_sugar import mf_flatten
from metricflow_semantics.experimental.metricflow_exception import MetricFlowInternalError
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.mf_graph.mf_graph import (
    MetricFlowGraph,
)
from metricflow_semantics.experimental.mf_graph.mutable_graph import EdgeT, NodeT
from metricflow_semantics.experimental.mf_graph.path_finding.graph_path import MutablePathT
from metricflow_semantics.experimental.mf_graph.path_finding.pathfinder_result import (
    FindAncestorsResult,
    FindDescendantsResult,
)
from metricflow_semantics.experimental.mf_graph.path_finding.traversal_profile import (
    GraphTraversalProfile,
    MutableGraphTraversalProfile,
)
from metricflow_semantics.experimental.mf_graph.path_finding.weight_function import WeightFunction
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet, OrderedSet
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


EdgeWeightFunctionT = TypeVar("EdgeWeightFunctionT", bound="WeightFunction")


class MetricFlowPathfinder(Generic[NodeT, EdgeT, MutablePathT], ABC):
    """Finds paths and related nodes (e.g. ancestors) via common graph algorithms.

    * This can be swapped out with another implementation.
    * `networkx` was evaluated, but there were some issues with the API (some methods did not support the required
      arguments).
    """

    _MAX_BFS_ITERATION_COUNT: Final[int] = 100

    def __init__(self) -> None:  # noqa: D107
        self._local_state = _MetricFlowPathfinderLocalState()
        self._lo = MutableGraphTraversalProfile()
        self._verbose_debug_logs = False

    def find_paths_dfs(
        self,
        graph: MetricFlowGraph[NodeT, EdgeT],
        initial_path: MutablePathT,
        target_nodes: Set[NodeT],
        weight_function: WeightFunction[NodeT, EdgeT, MutablePathT],
        max_path_weight: int,
        node_allow_set: Optional[Set[NodeT]] = None,
        node_deny_set: Optional[Set[NodeT]] = None,
        traversal_description: Optional[str] = None,
    ) -> Iterator[MutablePathT]:
        """Find paths (no cycles) to the target nodes via DFS.

        Args:
            graph: The graph to traverse.
            initial_path: The mutable path that should be extended during traversal.
            target_nodes: The nodes where the traversal should stop.
            weight_function: The weight function that should be used to compute the integer weight of a path. Edges can
            be blocked if the function returns `None`.
            max_path_weight: The maximum allowed weight of a path as computed by the provided weight function.
            node_allow_set: If specified, only these nodes will be considered for traversal. Otherwise, all nodes in the
            graph are allowed.
            node_deny_set: If specified, these nodes will be excluded from traversal.
            traversal_description: A description of the traversal for log / error messages.

        Returns: An iterator that returns all possible paths to any of the target nodes.
        """
        traversal: _DfsTraversal[NodeT, EdgeT, MutablePathT] = _DfsTraversal(
            graph=graph,
            initial_path=initial_path,
            target_nodes=target_nodes,
            weight_function=weight_function,
            max_path_weight=max_path_weight,
            node_allow_set=node_allow_set,
            node_deny_set=node_deny_set,
            traversal_profile=self._local_state.traversal_profile,
            traversal_description=traversal_description,
            verbose_debug_logs=self._verbose_debug_logs,
        )
        return traversal.find_paths()

    def find_descendants(
        self,
        graph: MetricFlowGraph[NodeT, EdgeT],
        source_nodes: OrderedSet[NodeT],
        target_nodes: OrderedSet[NodeT],
        node_allow_set: Optional[Set[NodeT]] = None,
        downward_closed: bool = False,
        deny_labels: Optional[Set[MetricFlowGraphLabel]] = None,
        max_iteration_count: int = _MAX_BFS_ITERATION_COUNT,
    ) -> FindDescendantsResult[NodeT]:
        """Find descendant nodes of the source nodes via BFS.

        Args:
            graph: The graph to traverse.
            source_nodes: Search for descendants of these nodes.
            target_nodes: When traversing the graph to find descendants, stop if any one of these nodes are reached.
            node_allow_set: If specified, only allow these nodes for traversal. Otherwise, all nodes in the graph are
            allowed.
            downward_closed: Whether to return descendants that are downward-closed with respect to the set of reachable
            nodes (source nodes + descendants).
            deny_labels: If specified, prevent traversal to edges or nodes with these labels.
            max_iteration_count: In each iteration, the descendants of a current set of nodes are added for visiting.
            This is the maximum number of iterations that should be done.

        Returns: A result object containing the descendants and other context.

        """
        # In each iteration, evaluate these nodes.
        batch_of_nodes_to_evaluate = list(source_nodes)
        # The target nodes that were reachable from at least one of the source nodes.
        reached_target_nodes: MutableOrderedSet[NodeT] = MutableOrderedSet()
        # All reachable nodes found during traversal.
        reachable_nodes: MutableOrderedSet[NodeT] = MutableOrderedSet(source_nodes)

        if self._verbose_debug_logs:
            logger.debug(LazyFormat("Finding descendants.", source_nodes=source_nodes, target_nodes=target_nodes))
        # Maps a descendant node to the source nodes that it is reachable from.
        node_to_reachable_source_nodes: dict[NodeT, MutableOrderedSet[NodeT]] = defaultdict(MutableOrderedSet)
        for node in source_nodes:
            node_to_reachable_source_nodes[node].add(node)

        iteration_index = 0
        labels_collected_during_traversal: set[MetricFlowGraphLabel] = set()
        labels_collected_during_traversal.update(*(source_node.labels for source_node in source_nodes))

        while iteration_index <= max_iteration_count:
            if len(batch_of_nodes_to_evaluate) == 0:
                break

            next_batch_of_node_to_evaluate: list[NodeT] = []

            examined_edge_count = 0
            for current_node in batch_of_nodes_to_evaluate:
                if current_node in target_nodes:
                    reached_target_nodes.add(current_node)
                    continue

                edges_to_examine = graph.edges_with_tail_node(current_node)

                for edge in edges_to_examine:
                    traversal_labels = edge.labels_for_path_addition
                    if deny_labels and (deny_labels & traversal_labels):
                        continue

                    head_node = edge.head_node
                    if node_allow_set is not None and head_node not in node_allow_set:
                        continue

                    if downward_closed and not all(
                        predecessor_of_head_node in reachable_nodes
                        for predecessor_of_head_node in graph.predecessors(head_node)
                    ):
                        continue

                    node_to_reachable_source_nodes[head_node].update(node_to_reachable_source_nodes[current_node])
                    reachable_nodes.add(head_node)
                    next_batch_of_node_to_evaluate.append(head_node)
                    labels_collected_during_traversal.update(traversal_labels)

                examined_edge_count += len(edges_to_examine)

            self._local_state.traversal_profile.increment_edge_examined_count(examined_edge_count)
            self._local_state.traversal_profile.increment_node_visit_count(len(batch_of_nodes_to_evaluate))
            batch_of_nodes_to_evaluate = next_batch_of_node_to_evaluate
            iteration_index += 1

        return FindDescendantsResult(
            reachable_nodes=reachable_nodes,
            reachable_target_nodes=reached_target_nodes,
            labels_collected_during_traversal=FrozenOrderedSet(sorted(labels_collected_during_traversal)),
            finish_iteration_index=iteration_index,
            target_node_to_reachable_source_nodes={
                target_node: node_to_reachable_source_nodes[target_node] for target_node in reached_target_nodes
            },
        )

    def find_ancestors(
        self,
        graph: MetricFlowGraph[NodeT, EdgeT],
        source_nodes: OrderedSet[NodeT],
        target_nodes: OrderedSet[NodeT],
        node_allow_set: Optional[Set[NodeT]] = None,
        deny_labels: Optional[Set[MetricFlowGraphLabel]] = None,
        upwards_closed: bool = False,
        max_iteration_count: int = _MAX_BFS_ITERATION_COUNT,
    ) -> FindAncestorsResult[NodeT]:
        """Find the ancestors of the source nodes.

        The arguments to this method are similar to the ones for `find_descendants`.
        """
        batch_of_nodes_to_evaluate: list[NodeT] = list(target_nodes)

        if self._verbose_debug_logs:
            logger.debug(LazyFormat("Finding ancestors.", source_nodes=source_nodes, target_nodes=target_nodes))

        iteration_index = 0

        labels_collected_during_traversal: set[MetricFlowGraphLabel] = set(
            mf_flatten(target_node.labels for target_node in target_nodes)
        )
        reachable_nodes: MutableOrderedSet[NodeT] = MutableOrderedSet(source_nodes)
        reachable_nodes.update(target_nodes)

        node_to_reachable_target_nodes: dict[NodeT, MutableOrderedSet[NodeT]] = defaultdict(MutableOrderedSet)

        for target_node in target_nodes:
            node_to_reachable_target_nodes[target_node].add(target_node)

        reachable_source_nodes: MutableOrderedSet[NodeT] = MutableOrderedSet()

        while iteration_index <= max_iteration_count:
            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Starting iteration",
                        reachable_nodes=batch_of_nodes_to_evaluate,
                    )
                )
            if len(batch_of_nodes_to_evaluate) == 0:
                break

            next_batch_of_node_to_evaluate: list[NodeT] = []
            examined_edges_count = 0
            for current_node in batch_of_nodes_to_evaluate:
                if current_node in source_nodes:
                    reachable_source_nodes.add(current_node)
                    continue

                predecessor_nodes_of_batch: Set[NodeT] = set(
                    mf_flatten(graph.predecessors(node) for node in batch_of_nodes_to_evaluate)
                )
                excluded_predecessor_nodes: set[NodeT] = set()
                for predecessor_node in predecessor_nodes_of_batch:
                    if node_allow_set is not None and predecessor_node not in node_allow_set:
                        if self._verbose_debug_logs:
                            logger.debug(
                                LazyFormat(
                                    "Excluding node as it's not in the allowed list", predecessor_node=predecessor_node
                                )
                            )
                        excluded_predecessor_nodes.add(predecessor_node)
                        continue

                    if deny_labels is not None and (deny_labels & predecessor_node.labels):
                        excluded_predecessor_nodes.add(predecessor_node)
                        if self._verbose_debug_logs:
                            logger.debug(
                                LazyFormat(
                                    "Excluding node as it matches the deny label set.",
                                    predecessor_node=predecessor_node,
                                    deny_labels=deny_labels,
                                )
                            )
                        continue

                    if upwards_closed and any(
                        successor_of_predecessor not in reachable_nodes
                        for successor_of_predecessor in graph.successors(predecessor_node)
                    ):
                        excluded_predecessor_nodes.add(predecessor_node)
                        if self._verbose_debug_logs:
                            logger.debug(
                                LazyFormat(
                                    "Excluding node as it's not upwards-closed.",
                                    predecessor_node=predecessor_node,
                                    successors=graph.successors(predecessor_node),
                                )
                            )
                        continue
                if self._verbose_debug_logs and excluded_predecessor_nodes:
                    logger.debug(
                        LazyFormat(
                            "Excluding selected predecessor nodes of this batch",
                            excluded_predecessor_nodes=excluded_predecessor_nodes,
                        )
                    )
                edges_to_examine = graph.edges_with_head_node(current_node)
                for edge_from_predecessor in graph.edges_with_head_node(current_node):
                    predecessor_node = edge_from_predecessor.tail_node
                    if predecessor_node in excluded_predecessor_nodes:
                        continue

                    next_batch_of_node_to_evaluate.append(predecessor_node)
                    labels_collected_during_traversal.update(
                        itertools.chain(edge_from_predecessor.labels, predecessor_node.labels)
                    )
                    reachable_nodes.add(predecessor_node)

                    node_to_reachable_target_nodes[predecessor_node].update(
                        node_to_reachable_target_nodes[edge_from_predecessor.head_node]
                    )
                examined_edges_count += len(edges_to_examine)

            self._local_state.traversal_profile.increment_edge_examined_count(examined_edges_count)
            self._local_state.traversal_profile.increment_node_visit_count(len(batch_of_nodes_to_evaluate))
            batch_of_nodes_to_evaluate = next_batch_of_node_to_evaluate
            iteration_index += 1

        return FindAncestorsResult(
            reachable_nodes=reachable_nodes,
            reachable_source_nodes=reachable_source_nodes,
            source_node_to_reachable_target_nodes={
                source_node: node_to_reachable_target_nodes[source_node] for source_node in reachable_source_nodes
            },
            labels_collected_during_traversal=FrozenOrderedSet(labels_collected_during_traversal),
            finish_iteration_index=iteration_index,
        )

    @property
    def traversal_profile_snapshot(self) -> GraphTraversalProfile:
        """Return a snapshot of the current counter set for pathfinder metrics.

        The counter set is used for logging / debugging.
        """
        return self._local_state.traversal_profile.copy()


class _DfsTraversal(Generic[NodeT, EdgeT, MutablePathT]):
    """A context object that contains shared state for a specific DFS traversal."""

    def __init__(
        self,
        graph: MetricFlowGraph[NodeT, EdgeT],
        initial_path: MutablePathT,
        target_nodes: Set[NodeT],
        weight_function: WeightFunction[NodeT, EdgeT, MutablePathT],
        max_path_weight: int,
        node_allow_set: Optional[Set[NodeT]],
        node_deny_set: Optional[Set[NodeT]],
        traversal_profile: MutableGraphTraversalProfile,
        traversal_description: Optional[str],
        verbose_debug_logs: bool,
    ) -> None:
        """See `find_paths_dfs` for description of the arguments."""
        if initial_path.is_empty:
            raise MetricFlowInternalError(
                LazyFormat(
                    "The initial path must have at least one node. Otherwise, there's no way to extend it.",
                    traversal_description=traversal_description,
                )
            )
        self._traversal_description = traversal_description
        self._current_path = initial_path
        self._graph = graph
        self._target_nodes = target_nodes
        self._weight_function = weight_function
        self._max_path_weight = max_path_weight
        self._node_allow_set = node_allow_set
        self._node_deny_set = node_deny_set
        self._traversal_profile = traversal_profile
        self._verbose_debug_logs = verbose_debug_logs

    def _get_valid_next_edges(self, current_node: NodeT) -> Sequence[tuple[EdgeT, int]]:
        """For a given node, figure out the next valid edges for the node."""
        current_weight = self._current_path.weight
        node_allow_set = self._node_allow_set
        node_deny_set = self._node_deny_set

        valid_next_edges: list[tuple[EdgeT, int]] = []
        candidate_edges = self._graph.edges_with_tail_node(current_node)
        for candidate_edge in candidate_edges:
            next_node = candidate_edge.head_node

            if node_allow_set is not None and next_node not in node_allow_set:
                continue

            if node_deny_set is not None and next_node in node_deny_set:
                continue

            # Block cycles.
            if candidate_edge.head_node in self._current_path.node_set:
                continue

            weight_added_by_candidate_edge = self._weight_function.incremental_weight(
                path_to_node=self._current_path,
                next_edge=candidate_edge,
            )

            if weight_added_by_candidate_edge is None:
                continue

            if current_weight + weight_added_by_candidate_edge > self._max_path_weight:
                continue

            valid_next_edges.append((candidate_edge, weight_added_by_candidate_edge))
        self._traversal_profile.increment_edge_examined_count(len(candidate_edges))
        return valid_next_edges

    def find_paths(self) -> Iterator[MutablePathT]:
        traversal_start_node = self._current_path.nodes[-1]
        return self._traverse_dfs(traversal_start_node)

    def _traverse_dfs(self, current_node: NodeT) -> Iterator[MutablePathT]:
        if self._verbose_debug_logs:
            logger.debug(LazyFormat("Visiting node", current_node=current_node, current_path=self._current_path))
        self._traversal_profile.increment_node_visit_count()
        current_path = self._current_path

        if current_node in self._target_nodes:
            self._traversal_profile.increment_generated_paths_count()
            yield current_path
            return

        for next_edge, incremental_weight in self._get_valid_next_edges(current_node):
            current_path.append_edge(next_edge, incremental_weight)
            for path in self._traverse_dfs(next_edge.head_node):
                yield path
            current_path.pop_end()


class _MetricFlowPathfinderLocalState(threading.local):
    def __init__(self) -> None:  # noqa: D107
        self.traversal_profile = MutableGraphTraversalProfile()
